/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2006 Bull S.A.S.
 * Contact: jonas-team@objectweb.org
 * Copyright (C) 2006 Distributed Systems Lab.
 * Universidad Polit??cnica de Madrid (Spain)
 * Contact: http://lsd.ls.fi.upm.es/lsd
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 *
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.lib.ejb21;

import java.rmi.RemoteException;
import java.util.Stack;

import javax.ejb.EJBException;

import org.ow2.jonas.lib.ejb21.ha.interceptor.HACurrentDelegateImpl;
import org.ow2.jonas.lib.util.Log;

import org.objectweb.carol.cmi.ObjectId;
import org.objectweb.carol.cmi.ServerConfigException;
import org.objectweb.carol.cmi.ServerIdFactory;
import org.objectweb.carol.cmi.ha.RequestId;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Generic part of the EJBObject implementation for replicated SFSBs
 * @author Francisco Perez-Sorrosal (fpsorrosal@no-spam@fi.upm.es)
 * @author Alberto Paz-Jimenez (apaz@no-spam@fi.upm.es)
 */
public abstract class JRepStatefulRemote extends JSessionRemote {

    /**
     * Logger for traces
     */
    protected static Logger cmilogger = Log.getLogger("org.objectweb.carol.cmi.ha");

    /**
     * Constructor
     * @param bf
     * @throws RemoteException
     */
    public JRepStatefulRemote(JSessionFactory bf) throws RemoteException {
        super(bf);
        cmilogger.log(BasicLevel.DEBUG, "");
    }

    /**
     * preInvoke is called before any request.
     * @param txa Transaction Attribute (Supports, Required, ...)
     * @return A RequestCtx object
     * @throws RemoteException Thrown when the method failed due to a
     *         system-level failure.
     */
    public RequestCtx preInvoke(int txa) throws RemoteException {
        RequestCtx rctx = super.preInvoke(txa);
        /* ******************REPLICATION CODE STARTS HERE*********************** */
        preInvokeHook(rctx); // Replication algorithm link
        /* **********************END OF REPLICATION CODE************************ */
        return rctx;
    }

    /**
     * postInvoke is called after any request.
     * @param rctx The RequestCtx that was returned at preInvoke()
     * @param remove TODO
     * @parem response The response that is going to be returned to the client
     *        (Needed by HA)
     * @throws RemoteException Thrown when the method failed due to a
     *         system-level failure.
     */
    public void postInvoke(RequestCtx rctx, Object response, boolean remove) throws RemoteException {
        /* ******************REPLICATION CODE STARTS HERE*********************** */
        try {
            postInvokeHook(rctx, response, remove); // Replication algorithm link
        /* **********************END OF REPLICATION CODE************************ */
        } catch (Exception e) {
            // COMPLETE: Handle errors
            e.printStackTrace();
            cmilogger.log(BasicLevel.ERROR, "Error calling postInvokeHook in sfsb.", e);
        } finally {
            super.postInvoke(rctx);
            // We dont need the request anymore so remove it
            removeCurrentRequest();
        }
    }

    /* ******************REPLICATION CODE STARTS HERE*********************** */

    // ------------------------------------------------------------------------
    // Protected methods ------------------------------------------------------
    // ------------------------------------------------------------------------
    /**
     * Injects the state of the bean if the cluster object id received in the
     * request is in the current node. It extracts the object id passed through
     * the interceptors from the request id included in the current HA context.
     */
    protected void injectState() {
        HACurrentDelegateImpl current = HACurrentDelegateImpl.getCurrent();
        // We still need the request so do not remove it
        Stack requests = current.getRequests();
        if (requests != null) {
            RequestId reqId = (RequestId) requests.peek();
            if (reqId != null) {
                JRepUtil.injectState(reqId.getObjectId(), (JStatefulSwitch) this.bs);
            }
        }
    }

    /**
     * Checks if there is a response associated with the current request id
     * @return true if the current rquest id have an associated response
     */
    protected boolean hasResponse() {
        HACurrentDelegateImpl current = HACurrentDelegateImpl.getCurrent();
        // We still need the request so do not remove it
        Stack requests = current.getRequests();
        if (requests != null) {
            RequestId reqId = (RequestId) requests.peek();
            if (reqId != null) {
                try {
                    return JRepUtil.hasBackupResponse(reqId);
                } catch (Exception e) {
                    cmilogger.log(BasicLevel.ERROR, "Cannot check response : " + reqId, e);
                    return false;
                }
            } else {
                return false;
            }
        } else {
            return false;
        }
    }

    /**
     * Obtain the response associated with the current request id.
     * Returns null if there is not response associated with the current
     * requestid
     * IMPORTANT: This method remove the request id from the requests heap
     * @return the response associated with the current request id
     */
    protected Object getResponse() {
        Object response = null;

        HACurrentDelegateImpl current = HACurrentDelegateImpl.getCurrent();
        // We still need the request so do not remove it
        Stack requests = current.getRequests();
        if (requests != null) {
            RequestId reqId = (RequestId) requests.peek();
            if (reqId != null) {
                try {
                    response = JRepUtil.getRepMgr().getBackupResponse(reqId);
                } catch (Exception e) {
                    cmilogger.log(BasicLevel.ERROR, "Cannot obtain response : " + reqId, e);
                    response = null;
                }
            } else {
                response = null;
            }

            // Remove the request
            requests.pop();
        }
        return response;
    }

    // ------------------------------------------------------------------------
    // Private methods --------------------------------------------------------
    // ------------------------------------------------------------------------

    /**
     * preInvokeHook implements the hook to perform the required replication
     * tasks before the bean context has been established. In this case, it
     * saves the current state of the bean before perform the invocation, in the
     * request context.
     * @param rctx The RequestCtx that was returned at preInvoke()
     * @throws RemoteException
     */
    private void preInvokeHook(RequestCtx rctx) throws RemoteException {
        try {
            // If the call is done from other JVM this is the root request
            HACurrentDelegateImpl current = HACurrentDelegateImpl.getCurrent();
            // We still need the request so do not remove it
            Stack requests = current.getRequests();
            if (requests != null) {
                RequestId reqId = (RequestId) requests.peek();
                // Discard external request ids
                if (!reqId.getObjectId().getServerId().equals(ServerIdFactory.getServerId())) {
                    requests.clear();
                    requests.push(reqId);
                }

                JRepUtil.preInvokeHook(bs, reqId.getObjectId(), rctx);
            }
        } catch (ServerConfigException sce) {
            cmilogger.log(BasicLevel.ERROR, "Unable to get Server Id.", sce);
        } catch (EJBException e) {
            e.printStackTrace();
            cmilogger.log(BasicLevel.ERROR, "Unable to get sfsb state.", e);
        }
    }

    /**
     * postInvokeHook implements the hook to perform the required replication
     * tasks after the bean context has been established
     * @param rctx The RequestCtx that was returned at preInvoke()
     * @parem response The response that is going to be returned to the client
     * @param remove true if it is an invocation to a remove method
     */
    private void postInvokeHook(RequestCtx rctx, Object response, boolean remove) {
        HACurrentDelegateImpl current = HACurrentDelegateImpl.getCurrent();
        RequestId reqId = null;
        ObjectId clusterOId = null;

        Stack requests = current.getRequests();
        if (requests != null) {
            reqId = (RequestId) requests.peek();
            clusterOId = reqId.getObjectId();

            JRepUtil.postInvokeHook(rctx, clusterOId, bs, remove);

            // If this is the root request, then associate response with rootId
            if (requests.size() == 1) {
                try {
                    JRepUtil.getRepMgr().addResponse(reqId, response);
                } catch (Exception e) {
                    e.printStackTrace();
                    cmilogger.log(BasicLevel.ERROR, "unable to obtain replication manager.", e);
                }
            }
        }
    }

    /**
     * Removes the curren request from the requests heap
     */
    private void removeCurrentRequest() {
        HACurrentDelegateImpl current = HACurrentDelegateImpl.getCurrent();
        Stack requests = current.getRequests();
        if (requests != null) {
            requests.pop();
        }
    }


    /* **********************END OF REPLICATION CODE************************ */
}
