/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id:  $
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.lib.jmbeans;

import java.util.Date;

import javax.management.JMException;
import javax.management.MBeanRegistration;
import javax.management.MBeanServer;
import javax.management.MBeanServerNotification;
import javax.management.Notification;
import javax.management.NotificationBroadcasterSupport;
import javax.management.NotificationListener;
import javax.management.ObjectName;

/**
 * The JSR77 J2EEServer MBean implementation.
 * Implement the StateManagement and Events specified in the JSR77.
 * @author Adriana.Danes@bull.net
 *
 */
public class J2EEServer extends NotificationBroadcasterSupport implements NotificationListener, MBeanRegistration {

    // === J2EEManagedObject == //
    /**
     * The managed object name
     */
    private String objectName;

    /**
     * State management support for this managed object (start, stop, ...)
     */
    private boolean stateManageable;

    /**
     * Performance statistics support for this managed object.
     */
    private boolean statisticsProvider;

    /**
     * Event provider support for this managed object.
     */
    private boolean eventProvider;

    // === EventProvider == //
    private final String [] standardTypes = {"j2ee.state.starting"
            , "j2ee.state.running"
            , "j2ee.state.stopped", "j2ee.state.failed"};

    private String [] eventTypes = null;

    // needed to send JMX notifications
    private long sequenceNumber = 0;
    private MBeanServer mbeanServer = null;

    // == StateManageable == //
    private final int starting = 0;
    private final int running = 1;
    private final int stopping = 2;
    private final int stopped = 3;
    private final int failed = 4;

    private int state;

    private long startTime;

    // == Static properties of the server == //
    private String serverVendor = null;
    private String serverVersion = null;
    private String serverName = null;
    private String domainName = null;
    private String jonasBase = null;
    private String protocols = null;

    /**
     * MBean constructor.
     * @param objectName object name of the managed object
     */
    public J2EEServer(String objectName) {
        super();
        this.objectName = objectName;
        this.stateManageable = false;
        this.statisticsProvider = false;
        this.eventProvider = false;
    }

    /**
     * MBean constructor.
     * @param objectName object name of the managed object
     * @param stateManageable if true, this managed object implements J2EE State
     *        Management Model
     * @param statisticsProvider if true, this managed object implements the
     *        J2EE StatisticProvide Model
     * @param eventProvider if true, this managed object implements the J2EE
     *        EventProvider Model
     */
    public J2EEServer(String objectName, boolean stateManageable, boolean statisticsProvider,
            boolean eventProvider) {
        this.objectName = objectName;
        this.stateManageable = stateManageable;
        this.statisticsProvider = statisticsProvider;
        this.eventProvider = eventProvider;
        if (eventProvider) {
            eventTypes = standardTypes;
        }
        if (stateManageable) {
            Date d = new Date();
            startTime = d.getTime();
        }
        setStarting();
    }

    // MBean interface implementation
    public String getObjectName() {
        return objectName;
    }

    /**
     * @return true if it is an event provider
     */
    public boolean isEventProvider() {
        return eventProvider;
    }

    /**
     * @return true if this managed object implements J2EE State Management
     *         Model
     */
    public boolean isStateManageable() {
        return stateManageable;
    }

    /**
     * @return true if this managed object implements the J2EE StatisticProvider
     *         Model
     */
    public boolean isStatisticsProvider() {
        return statisticsProvider;
    }

    public String[] getEventTypes() {
        return eventTypes;
    }

    public int getState() {
        return state;
    }

    public long getStartTime() {
        return startTime;
    }

    public String getServerVendor() {
        return serverVendor;
    }

    public void setServerVendor(String serverVendor) {
        this.serverVendor = serverVendor;
    }

    public String getServerVersion() {
        return serverVersion;
    }

    public void setServerVersion(String serverVersion) {
        this.serverVersion = serverVersion;
    }

    public String getServerName() {
        return serverName;
    }

    public void setServerName(String serverName) {
        this.serverName = serverName;
    }

    public String getJonasBase() {
        return jonasBase;
    }

    public void setJonasBase(String jonasBase) {
        this.jonasBase = jonasBase;
    }

    public String getProtocols() {
        return protocols;
    }

    public void setProtocols(String protocols) {
        this.protocols = protocols;
    }

    public String[] getDeployedObjects() {
        String[] result = new String[0];
        return result;
    }

    public String getDomainName() {
        return domainName;
    }

    public void setDomainName(String domainName) {
        this.domainName = domainName;
    }

    public void start() {
        if (state == stopped || state == failed) {
            // start operation is allowed
            setStarting();
            // TODO do something in order to make the framework start the required services
            System.out.println("WARNING - this operation is not yet implemented");
        } else {
            // start operation isnot allowed
            throw new IllegalStateException("The start() operation can be invoked only when the server is in STOPPED or FAILED state");
        }
    }

    public void startRecursive() {
        if (state == stopped || state == failed) {
            // start operation is allowed
            setStarting();
            // TODO do something in order to make the framework start the required services
            System.out.println("WARNING: start JOnAS operation is not yet implemented!");
        } else {
            // start operation isnot allowed
            throw new IllegalStateException("The startRecursive() operation can be invoked only when the server is in STOPPED or FAILED state");
        }
    }

    public void stop() {
        if (state == running || state == starting) {
            // stop operation is allowed
            setStopping();
            // TODO do something in order to make the framework stop the services
            System.out.println("WARNING: stop JOnAS operation is not yet implemented!");
        } else if (state == failed) {
            // cf. to the JSR77, may try to stop in a FAILED state
            setStopping();
            // TODO do something in order to make the framework stop the services
            System.out.println("WARNING: stop JOnAS operation is not yet implemented!");
        } else {
            // start operation isnot allowed
            throw new IllegalStateException("The startRecursive() operation can be invoked only when the server is in STOPPED or FAILED state");
        }
    }

    // Other JMX methods
    public void handleNotification(Notification notification, Object handback) {
        System.out.println("Received notification of type: " + notification.getType());
        if (notification instanceof MBeanServerNotification) {
            // Treat notifications emitted by the jmx server
            // TODO
        } else {
            // Treat notifications emitted by the OsgiMonitor MBean
            // Implement the JSR77 StateManagement object state diagram
            if (notification.getType().equals(JNotification.startingSuccess)) {
                if (state == starting) {
                    setRunning();
                }
            }
            if (notification.getType().equals(JNotification.startingFail)) {
                if (state == starting) {
                    setRunning();
                }
            }
            if (notification.getType().equals(JNotification.stoppingSuccess)) {
                if (state == stopping) {
                    setStopped();
                }
            }
            if (notification.getType().equals(JNotification.stoppingFail)) {
                if (state == stopping) {
                    setFailed();
                }
            }
            if (notification.getType().equals(JNotification.serverFail)) {
                if (state == running) {
                    setFailed();
                }
            }
        }
    }

    public void postDeregister() {
    }

    public void postRegister(Boolean registrationDone) {
        try {
            ObjectName on = ObjectName.getInstance("JMImplementation:type=MBeanServerDelegate");
            mbeanServer.addNotificationListener(on, this, null, null);
            on = ObjectName.getInstance(domainName + ":type=OsgiMonitor");
            mbeanServer.addNotificationListener(on, this, null, null);
            System.out.println("Registered J2EEServer MBean (" + getObjectName() + ") with state = " + getState());
        } catch (JMException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (NullPointerException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }

    public void preDeregister() throws Exception {
    }

    public ObjectName preRegister(MBeanServer server, ObjectName name) throws Exception {
        this.mbeanServer = server;
        if (name == null) {
            return ObjectName.getInstance(objectName);
        } else {
            return name;
        }
    }

    // Private methods for state managemrnt
    private void setStarting() {
        state = starting;
        JNotification notif = new JNotification(JNotification.startingType
                , getObjectName(), sequenceNumber++);
    }

    private void setRunning() {
        state = running;
        JNotification notif = new JNotification(JNotification.runningType
                , getObjectName(), sequenceNumber++);
        sendNotification(notif);
    }

    private void setStopping() {
        state = stopping;
        JNotification notif = new JNotification(JNotification.stoppingType
                , getObjectName(), sequenceNumber++);
        sendNotification(notif);
        // Temporary - allows to test start() invocation after a stop()
        // setStopped();
    }

    private void setStopped() {
        state = stopped;
        JNotification notif = new JNotification(JNotification.stoppedType
                , getObjectName(), sequenceNumber++);
        sendNotification(notif);
    }

    private void setFailed() {
        state = failed;
        JNotification notif = new JNotification(JNotification.failedType
                , getObjectName(), sequenceNumber++);
        sendNotification(notif);
    }
}