/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.lib.jmbeans.osgi;
import java.util.Iterator;
import java.util.List;

import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;

import org.osgi.framework.BundleActivator;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceReference;
import org.ow2.jonas.Version;
import org.ow2.jonas.jmx.JmxService;
import org.ow2.jonas.lib.jmbeans.J2EEServer;
import org.ow2.jonas.lib.jmbeans.osgi.monitoring.OsgiMonitor;
import org.ow2.jonas.properties.ServerProperties;
import org.ow2.jonas.registry.RegistryService;

/**
 * Activator for the bundle which creates the basic JSR77 MBeans
 * @author Adriana Danes
 */
public class Activator implements BundleActivator {
    /**
     * jmx service instance
     */
    private JmxService jmx = null;

    /**
     * registery service instance
     */
    private RegistryService registry = null;
    /**
     * Basic naming properties:
     * - the management domain name
     */
    private String domainName = null;
    /**
     * - the server instance name
     */
    private String serverName = null;

    public void start(final BundleContext bc) throws Exception {
        System.out.println("Starting jonas-mbeans Bundle...");
        // Get domainName and serverName from the Bootstrap service
        // TODO assume service already started.
        ServiceReference sr = bc.getServiceReference(ServerProperties.class.getName());
        ServerProperties props = null;
        if (sr != null) {
            props = (ServerProperties) bc.getService(sr);
        } else {
            System.out.println("Bootstrap Bundle not running");
            System.out.println("Failed to start jonas-mbeans Bundle");
            return;
        }
        // Need the jmx service to register MBeans
        sr = bc.getServiceReference(JmxService.class.getName());
        if (sr != null) {
            jmx = (JmxService) bc.getService(sr);
        } else {
            System.out.println("JmxService not running");
            System.out.println("Failed to start jonas-mbeans Bundle");
            return;
        }
        // Need the registry service to get the protocols list from it
        sr = bc.getServiceReference(RegistryService.class.getName());
        if (sr != null) {
            registry = (RegistryService) bc.getService(sr);
        } else {
            System.out.println("RegistryService not running");
            System.out.println("Failed to start jonas-mbeans Bundle");
            return;
        }

        // Test we got properties necessary to create the MBeans
        if (props == null) {
            System.out.println("Couldn't get all the required properties to register JSR77 MBeans");
            throw new Exception("Couldn't get all the required properties to register JSR77 MBeans");
        }

        // Got all we need to create and register JSR77 MBeans
        try {
            domainName = props.getDomainName();
            serverName = props.getServerName();
            registerMBeans(props);
            // unset props as we don't need properties any more
            props = null;
        } catch (MalformedObjectNameException me) {
            System.out.println("Failed to register mbeans: " + me.getMessage());
            return;
        } catch (Exception e) {
            System.out.println("Failed to register mbeans");
            e.printStackTrace();
            return;
        }
    }

    public void stop(final BundleContext bc) throws Exception {
        System.out.println("Stopping jonas-mbeans Bundle...");
        unregisterMBeans(domainName, serverName);
        jmx = null;
        registry = null;
    }

    /**
     * Register JSR77 MBeans
     * @param domainName
     * @param serverName
     * @param props
     * @throws Exception
     */
    private void registerMBeans(ServerProperties props) throws Exception {
        // Create and register OsgiMonitor MBean
        String name = domainName + ":type=OsgiMonitor";
        ObjectName objectName = ObjectName.getInstance(name);
        System.out.println("Registere MBean: " + objectName.toString());
        OsgiMonitor mbean1 = new OsgiMonitor(objectName.toString());
        jmx.registerMBean(mbean1, objectName);
        System.out.println("MBeans registered: " + objectName.toString());

        // Create and register J2EEServer MBean
        name = domainName + ":j2eeType=J2EEServer,name=" + serverName;
        objectName = ObjectName.getInstance(name);
        J2EEServer mbean = new J2EEServer(objectName.toString(), true, false, true);
        mbean.setServerVendor(Version.VENDOR);
        mbean.setServerVersion(Version.getNumber());
        mbean.setServerName(serverName);
        mbean.setDomainName(domainName);
        mbean.setJonasBase(props.getValue("jonas.base"));
        mbean.setProtocols(getProtocols());
        jmx.registerModelMBean(mbean, objectName);
        System.out.println("MBeans registered: " + objectName.toString());

        // Temporary
        mbean1.temporary();
    }

    private void unregisterMBeans(String domainName, String serverName) throws Exception {
        // Unregister 2EEServer MBean
        String name = domainName + ":j2eeType=J2EEServer,name=" + serverName;
        ObjectName objectName = ObjectName.getInstance(name);
        jmx.unregisterModelMBean(objectName);
        System.out.println("MBeans unregistered: " + objectName.toString());

        // Unregister OsgiMonitor MBean
        name = domainName + ":type=OsgiMonitor";
        objectName = ObjectName.getInstance(name);
        jmx.unregisterModelMBean(objectName);
        System.out.println("MBeans unregistered: " + objectName.toString());
    }

    /**
     * @return Get the protocols list from the registry service
     */
    private String getProtocols() {
        String protocols = "";
        String prefix = "rmi/";
        List protocolList = registry.getActiveProtocolNames();
        Iterator it = protocolList.iterator();
        while (it.hasNext()) {
            String protocolName = (String) it.next();
            if (protocols == "") {
                protocols = prefix + protocolName;
            } else {
                protocols += "," +  prefix + protocolName;
            }
        }
        return protocols;
    }

}
