/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.lib.naming.osgi;

import java.util.Hashtable;

import javax.naming.Context;
import javax.naming.NamingException;
import javax.naming.NoInitialContextException;
import javax.naming.spi.InitialContextFactory;
import javax.naming.spi.InitialContextFactoryBuilder;
import javax.naming.spi.NamingManager;

import org.osgi.framework.BundleContext;

/**
 * Implementation of {@link InitialContextFactoryBuilder} that knows
 * how to load classes in an OSGi environment.
 *
 * It try to load the {@link InitialContextFactory} implementation
 * from various classloaders:
 * <ol>
 *   <li>Thread currentThread context ClassLoader.</li>
 *   <li>The Bundle ClassLoader that have loaded this class.</li>
 *   <li>A {@link ClassLoader} provided in the environment (<code>JONAS_NAMING_LOADER</code>)</li>
 * </ol>
 *
 * @author Guillaume Sauthier
 */
public class OSGiInitialContextFactoryBuilder implements InitialContextFactoryBuilder {

    /**
     * Property key.
     */
    public static final String JONAS_NAMING_LOADER = "org.ow2.jonas.naming.osgi.loader";

    /**
     * The {@link BundleContext} to use for class loading.
     */
    private BundleContext bundleContext;

    public OSGiInitialContextFactoryBuilder(BundleContext bc) {
        this.bundleContext = bc;

        // Set our own InitialContextFactoryBuilder
        // Try to load ICF classes from various ClassLoaders
        if (!NamingManager.hasInitialContextFactoryBuilder()) {
            // Only set a new Builder if none was specified
            try {
                NamingManager.setInitialContextFactoryBuilder(this);
            } catch (NamingException e) {
                e.printStackTrace();
            }
        }
    }

    /**
     * Load the {@link InitialContextFactory} instances from:
     * <ol>
     *   <li>Thread currentThread context ClassLoader.</li>
     *   <li>the Bundle ClassLoader that have loaded this class.</li>
     *   <li>a ClassLoader provided in the environment (<code>JONAS_NAMING_LOADER</code>)</li>
     * </ol>
     * @see javax.naming.spi.InitialContextFactoryBuilder#createInitialContextFactory(java.util.Hashtable)
     * @see OSGiInitialContextFactoryBuilder.JONAS_NAMING_LOADER
     */
    @SuppressWarnings("unchecked")
    public InitialContextFactory createInitialContextFactory(Hashtable<?, ?> env) throws NamingException {

        String className = null;
        if (env != null) {
            className = (String) env.get(Context.INITIAL_CONTEXT_FACTORY);
        }
        if (className == null) {
            NoInitialContextException ne = new NoInitialContextException(
                "Need to specify class name in environment or system " +
                "property, or as an applet parameter, or in an " +
                "application resource file:  " + Context.INITIAL_CONTEXT_FACTORY);
            throw ne;
        }

        // 1. Try Thread currentThread context ClassLoader
        ClassLoader threadLoader = Thread.currentThread().getContextClassLoader();
        try {
            return getInitialContextFactoryInstance(className, threadLoader);
        } catch(Exception e) {
        }

        // 2. Try from this bundle loader
        try {
            Class<? extends InitialContextFactory> clazz = bundleContext.getBundle().loadClass(className)
                                                            .asSubclass(InitialContextFactory.class);
            return clazz.newInstance();
        } catch(Exception e) {
        }

        // 3. Try from a provided ClassLoader
        ClassLoader providedLoader = (ClassLoader) env.get(JONAS_NAMING_LOADER);
        if (providedLoader != null) {
            try {
                return getInitialContextFactoryInstance(className, providedLoader);
            } catch(Exception e) {
            }
        }

        // If we were unable to load the class
        // Build a comprehensible message
        StringBuilder sb = new StringBuilder();
        sb.append("Cannot instantiate class: '" + className + "'. ");
        sb.append("\n * Was not found in the Thread Context ClassLoader: " + threadLoader + ".");
        sb.append("\n * Was not found from the ClassLoader of the bundle: " + bundleContext.getBundle().getSymbolicName() + ".");
        if (providedLoader != null) {
            sb.append("\n * Was not found in the provided ClassLoader: " + providedLoader + ".");
        }

        // and throw the Exception
        NoInitialContextException ne =  new NoInitialContextException(sb.toString());
        throw ne;
    }

    /**
     * @param className {@link InitialContextFactory} implmentation class name.
     * @param loader The loader to try for class loading.
     * @return An {@link InitialContextFactory} instance.
     * @throws ClassNotFoundException Class was not found in the ClassLoader.
     * @throws InstantiationException Couldn't create a new ICF instance.
     * @throws IllegalAccessException Instantiation error
     */
    private InitialContextFactory getInitialContextFactoryInstance(String className, ClassLoader loader) throws ClassNotFoundException,
            InstantiationException, IllegalAccessException {
        // Try to load the class from the given loader
        Class<? extends InitialContextFactory> clazz = Class.forName(className,
                                                                     true,
                                                                     loader)
                                                            .asSubclass(InitialContextFactory.class);
        return clazz.newInstance();
    }

}
