/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2006-2007 Bull S.A.S.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.lib.management.domain;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Vector;

import javax.management.JMException;
import javax.management.MBeanServer;
import javax.management.MBeanServerConnection;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.jonas.deployment.domain.DomainMap;
import org.ow2.jonas.deployment.domain.DomainMapException;
import org.ow2.jonas.deployment.domain.wrapper.DomainManagerWrapper;
import org.ow2.jonas.deployment.domain.xml.Cluster;
import org.ow2.jonas.deployment.domain.xml.ClusterDaemon;
import org.ow2.jonas.deployment.domain.xml.Location;
import org.ow2.jonas.deployment.domain.xml.Server;
import org.ow2.jonas.discovery.DiscoveryEvent;
import org.ow2.jonas.discovery.DiscoveryState;
import org.ow2.jonas.jmx.JmxService;
import org.ow2.jonas.lib.management.domain.cluster.BaseCluster;
import org.ow2.jonas.lib.management.domain.cluster.ClusterFactory;
import org.ow2.jonas.lib.management.domain.cluster.cmi.CmiClusterFactory;
import org.ow2.jonas.lib.management.domain.cluster.ha.EjbHaClusterFactory;
import org.ow2.jonas.lib.management.domain.cluster.jk.JkClusterFactory;
import org.ow2.jonas.lib.management.domain.cluster.logical.LogicalCluster;
import org.ow2.jonas.lib.management.domain.cluster.logical.LogicalClusterFactory;
import org.ow2.jonas.lib.management.domain.cluster.tomcat.TomcatClusterFactory;
import org.ow2.jonas.lib.management.domain.proxy.clusterd.ClusterDaemonProxy;
import org.ow2.jonas.lib.management.domain.proxy.server.ServerProxy;
import org.ow2.jonas.lib.util.JonasObjectName;
import org.ow2.jonas.lib.util.Log;
import org.ow2.jonas.service.manager.ServiceManager;

/**
 * This class represents the Master Server of the Domain.
 * It manages the servers and the clusters in the domain.
 * @author danesa
 * @author durieuxp refactoring
 */
public class DomainMonitor {

    /**
     * logger for traces
     */
    private static Logger logger = Log.getLogger("org.ow2.jonas.management.domain");

    /**
     * List of known ClusterFactory
     */
    private ArrayList clusterFactoryList = new ArrayList(5);

    /**
     * List of known ClusterDaemon
     */
    private HashMap myclusterdaemons = new HashMap();

    /**
     * ref to the Jmx Service
     */
    private JmxService jmx = (JmxService) ServiceManager.getInstance().getJmxService();

    /**
     * ref to the MBeanServer.
     */
    private MBeanServer mbeanServer;

    /**
     * Thread object allowing to check the proxys state in order to detect if
     * they can be reached using the established connection.
     */
    private StateMonitor stateMonitor = null;

    /**
     * My Domain Name
     */
    private String domainName;

    /**
     * Local Server Name
     */
    private String jonasServerName;

    /**
     * The local server name (name of the master)
     */
    private String masterName;

    /**
     * Domain description string (from domain.xml)
     */
    private String description;

    /**
     * list of deployement operations
     */
    private ArrayList deployList = new ArrayList();

    /**
     * removed server by cluster daemons
     */
    private HashMap<String, ArrayList<String>> removedServerByClusterd = new HashMap<String, ArrayList<String>>();
    /**
     * The LogicalClusterFactory
     */
    private LogicalClusterFactory lcl;

    private boolean ismaster = false;

    private static DomainMonitor unique;

    /**
     * Constructor
     */
    public DomainMonitor(String domain) {
        domainName = domain;
        mbeanServer = jmx.getJmxServer();
        unique = this;
        jonasServerName = jmx.getJonasServerName();
    }

    /**
     * Singleton: Each server (master or slave) must have 1 unique DomainMonitor object.
     * @return the unique instance of the DomainMonitor
     */
    public static DomainMonitor getInstance() {
        return unique;
    }

    public void setMaster() {
        ismaster = true;
        // Create the ClusterFactory objects
        lcl = new LogicalClusterFactory(this);
        clusterFactoryList.add(lcl);
        clusterFactoryList.add(new JkClusterFactory(this));
        clusterFactoryList.add(new TomcatClusterFactory(this));
        clusterFactoryList.add(new CmiClusterFactory(this));
        clusterFactoryList.add(new EjbHaClusterFactory(this));
    }

    public boolean isMaster() {
        return ismaster;
    }

    public void readDomainConfig() {

        // Create domain map based on the info contained in domain.xml file.
        try {
            createDomainMap();
        } catch (Exception e) {
            logger.log(BasicLevel.INFO, "Cannot read domain.xml");
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "Exception raised:" + e);
            }
        }

        // Create state monitor thread
        stateMonitor = new StateMonitor(this);
        stateMonitor.start();
        logger.log(BasicLevel.DEBUG, "State monitor created");
    }

    public String getDescription() {
        return description;
    }

    public String getDomainName() {
        return domainName;
    }

    public void setMonitoringPeriod(int sec) {
        stateMonitor.setSamplingPeriod(sec);
    }

    public int getMonitoringPeriod() {
         return stateMonitor.getSamplingPeriod();
    }

    /**
     * Create the domain representation found in domain.xml
     * @throws DomainMapException Error reading domain.xml
     * @throws JMException
     */
    private void createDomainMap() throws DomainMapException, JMException {

        // Parse the domain.xml file.
        ClassLoader currentLoader = Thread.currentThread().getContextClassLoader();
        DomainMap domainMap = DomainManagerWrapper.getDomainMap(null, currentLoader);
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "Domain Map :"  + domainMap);
        }

        // Check domain name.
        String definedName =  domainMap.getName();
        if (! domainName.equals(definedName)) {
            logger.log(BasicLevel.INFO, "Domain name in domain.xml is not the actual domain name: " + domainName);
            return;
        }

        description = domainMap.getDescription();

        // Get list of clusterdaemons
        Vector daemons = domainMap.getClusterDaemons();
        for (int i = 0; i < daemons.size(); i++) {
            ClusterDaemon cd = (ClusterDaemon) daemons.elementAt(i);
            String cdname = cd.getName();
            // Create the proxy and add it to the list
            Location loc = cd.getLocation();
            ClusterDaemonProxy cdp = new ClusterDaemonProxy(this, cdname, loc.getUrlList());
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "Adding clusterdaemon : "  + cdname);
            }
            // give a name and register mbean
            ObjectName on = JonasObjectName.clusterDaemonProxy(domainName, cdname);
            cdp.setObjectName(on.toString());
            // add it to the list
            myclusterdaemons.put(cdname, cdp);
        }

        // Get list of servers
        Vector servers = domainMap.getServers();
        for (int i = 0; i < servers.size(); i++) {
            Server sv = (Server) servers.elementAt(i);
            String svname = sv.getName();
            // Check ClusterDaemon attached to this server
            String cdn = sv.getClusterDaemon();
            ClusterDaemonProxy cdp = null;
            if (cdn != null) {
                cdp = (ClusterDaemonProxy) myclusterdaemons.get(cdn);
                if (cdp == null) {
                    logger.log(BasicLevel.WARN, "Undeclared ClusterDaemon :"  + cdn);
                }
            }
            // Create the proxy and add it to the list
            Location loc = sv.getLocation();
            ServerProxy sp = findServerProxy(svname);
            if (sp == null) {
                sp = new ServerProxy(this, svname, loc.getUrlList(), cdp);
                if (logger.isLoggable(BasicLevel.DEBUG)) {
                    logger.log(BasicLevel.DEBUG, "Adding server : "  + svname);
                }
                // give a name and register mbean
                ObjectName on = JonasObjectName.serverProxy(domainName, svname);
                sp.setObjectName(on.toString());
                // add it to the list
                lcl.notifyServer(sp);
            } else {
                logger.log(BasicLevel.WARN, "Server already declared: " + svname);
            }
        }

        // Get list of logical clusters
        Vector clusters = domainMap.getClusters();
        for (int i = 0; i < clusters.size(); i++) {
            Cluster cl = (Cluster) clusters.elementAt(i);
            String clname = cl.getName();
            LogicalCluster locl = (LogicalCluster) lcl.getCluster(clname);
            if (locl == null) {
                // Create a Logical Cluster
                locl = createLogicalCluster(clname);
            }
            if (locl != null) {
                // Add servers to the new cluster
                LinkedList serverList = cl.getServerList();
                for (int j = 0; j < serverList.size(); j++) {
                    Server sv = (Server) serverList.get(j);
                    String svname = sv.getName();
                    ServerProxy sp = findServerProxy(svname);
                    if (sp == null) {
                        // Server is defined directly in the Cluster
                        String cdn = sv.getClusterDaemon();
                        ClusterDaemonProxy cdp = null;
                        if (cdn != null) {
                            cdp = (ClusterDaemonProxy) myclusterdaemons.get(cdn);
                            if (cdp == null) {
                                logger.log(BasicLevel.WARN, "Undeclared ClusterDaemon :"  + cdn);
                            }
                        }
                        // Create the proxy and add it to the list
                        Location loc = sv.getLocation();
                        sp = new ServerProxy(this, svname, loc.getUrlList(), cdp);
                        // give a name and register mbean
                        ObjectName on = JonasObjectName.serverProxy(domainName, svname);
                        sp.setObjectName(on.toString());
                    }
                    locl.addServer(svname, sp);
                    // Notify the cluster factories
                    notifyServerProxyRunning(sp);
                }
            }
        }

    }

    /**
     * Add the local J2EEServer to the list of servers
     */
    public ServerProxy addLocalServer(String svname, Collection urls) {
        logger.log(BasicLevel.DEBUG, svname);

        // Add a ServerProxy even for the local server
        ServerProxy sp = findServerProxy(svname);
        if (sp == null) {
            sp = new ServerProxy(this, svname, urls, null);
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "Adding server : "  + svname);
            }
            // give a name and register mbean
            ObjectName on;
            try {
                on = JonasObjectName.serverProxy(domainName, svname);
                sp.setObjectName(on.toString());
            } catch (MalformedObjectNameException e) {
                logger.log(BasicLevel.ERROR, "MalformedObjectName: "  + svname);
            }
            // add it to the list
            lcl.notifyServer(sp);
            // put monitoring info in the ServerProxy
            sp.getMonitoringInfo();
        }
        // why here ? - to be changed ?
        readDomainConfig();

        masterName = svname;
        return sp;
    }

    /**
     * Get a Cluster by its name. Look in all cluster factories.
     * @param name The name of the cluster
     * @return the Cluster or null if unknown by the domain.
     */
    public BaseCluster findCluster(String name) {
        logger.log(BasicLevel.DEBUG, name);
        if (! ismaster) {
            throw new RuntimeException("This is not the master.");
        }
        BaseCluster cluster = null;
        // Look in each cluster factory to find if this cluster exists somewhere
        for (Iterator it = clusterFactoryList.iterator(); it.hasNext();) {
            ClusterFactory cf = (ClusterFactory) it.next();
            cluster = cf.getCluster(name);
            if (cluster != null) {
                break;
            }
        }
        return cluster;
    }

    /**
     * Get a Server by its name
     * @param name The name of the server
     * @return the ServerProxy or null if unknown by the domain.
     */
    public ServerProxy findServerProxy(String name) {
        logger.log(BasicLevel.DEBUG, name);
        if (! ismaster) {
            throw new RuntimeException("This is not the master.");
        }
        // Look in the domain cluster, holding all servers in the domain.
        BaseCluster cluster = (BaseCluster) lcl.getCluster(domainName);
        if (cluster == null) {
            // Not initialized yet: no server known at this time.
            return null;
        }
        return cluster.getServerProxy(name);
    }

    /**
     * Get a ClusterDaemon by its name
     * @param name The name of the clusterdaemon
     * @return the ClusterDaemonProxy or null if unknown by the domain.
     */
    public ClusterDaemonProxy findClusterDaemonProxy(String name) {
        if (! ismaster) {
            throw new RuntimeException("This is not the master.");
        }
        //the return value will be null if the clusterd si not yet registered
        return  (ClusterDaemonProxy) myclusterdaemons.get(name);
    }
    /**
     * register a cluster daemon proxy in the domain
     * @param cdName the cluster daemon's name
     * @param serverName name of the current jonas server (master instance)
     * @throws MalformedObjectNameException
     */

    @SuppressWarnings({ "unchecked", "unchecked" })
    public void registerClusterDaemonProxy(String cdName, String serverName, Collection urls ) throws MalformedObjectNameException {
        if (! ismaster) {
            throw new RuntimeException("This is not the master.");
        }
             //this cluster daemon is not registered
             // give a name and register mbean
             ObjectName on = JonasObjectName.clusterDaemonProxy(domainName, cdName);
             ClusterDaemonProxy cdp = new ClusterDaemonProxy(this,cdName,urls);
             cdp.setObjectName(on.toString());
             // add it to the list
         if (myclusterdaemons.get(cdName) == null) {
             myclusterdaemons.put(cdName, cdp);
         } else {
             //the clusterd is already registered.
             //remove it first
             myclusterdaemons.remove(cdName);
             myclusterdaemons.put(cdName, cdp);
        }
    }

    /**
     * Get the list of ServerProxy
     * @return Collection of all ServerProxy in the Domain
     */
    public Collection getServerList() {

        if (! ismaster) {
            throw new RuntimeException("This is not the master.");
        }
        // All servers are set in the Domain Cluster
        LogicalCluster domaincluster = (LogicalCluster) lcl.getCluster(domainName);
        if (domaincluster == null) {
            return new ArrayList();
        }
        return domaincluster.getServerProxyList();
    }

    /**
     * Get the list of all ClusterDaemonProxy
     */
    public Collection getClusterDaemonList() {
        if (! ismaster) {
            throw new RuntimeException("This is not the master.");
        }
        return myclusterdaemons.values();
    }

    /**
     * Get the list of all clusters of any type
     * @return Collection of all Clusters in the domain
     */
    public Collection getTotalClusterList() {
        if (! ismaster) {
            throw new RuntimeException("This is not the master.");
        }
        ArrayList ret = new ArrayList();
        for (Iterator it = clusterFactoryList.iterator(); it.hasNext();) {
            ClusterFactory cf = (ClusterFactory) it.next();
            ret.addAll(cf.getClusterList());
        }
        return ret;
    }

    /**
     * Create a logical cluster
     */
    private LogicalCluster createLogicalCluster(String name) {
        if (! ismaster) {
            throw new RuntimeException("This is not the master.");
        }
        return lcl.createLogicalCluster(name);
    }

    /**
     * Get the list of logical clusters
     */
    public Collection getLogicalClusterList() {
        if (! ismaster) {
            throw new RuntimeException("This is not the master.");
        }
        return lcl.getClusterList();
    }


    /**
     * MBean method
     * To be replaced by J2EEDomain.getServers()
     * @return Array of OBJECT_NAMEs of all ServerProxy
     */
    public String[] getProxys() {
        if (! ismaster) {
            throw new RuntimeException("This is not the master.");
        }
        // Look in the domain cluster, holding all servers in the domain.
        BaseCluster cluster = (BaseCluster) lcl.getCluster(domainName);
        Collection col = cluster.getServerProxyList();
        String [] ret = new String[col.size()];
        int i = 0;
        for (Iterator it = col.iterator(); it.hasNext(); ) {
            ServerProxy proxy = (ServerProxy) it.next();
            ret[i++] = proxy.getObjectName();
        }
        return ret;
    }

    /**
     * MBean method
     * @return Array of OBJECT_NAMEs of all cluster MBeans
     */
    public String[] getClusters() {
        if (! ismaster) {
            throw new RuntimeException("This is not the master.");
        }
        ArrayList al = new ArrayList();
        for (int i = 0; i < clusterFactoryList.size(); i++) {
            ClusterFactory clFactory = (ClusterFactory) clusterFactoryList.get(i);
            Iterator clIt = clFactory.getClusterList().iterator();
            for (; clIt.hasNext();) {
                BaseCluster cl = (BaseCluster) clIt.next();
                al.add(cl.getObjectName());
            }
        }
        String [] ret = new String[al.size()];
        for (int i = 0; i < al.size(); i++) {
            ret[i] = (String) al.get(i);
        }
        return ret;
        /*
         *  This code was used in J2EEDomain.getClusters() method
         *  Its a-priori equivalent with the above code ....
            Collection col = getTotalClusterList();
            sb = new String[col.size()];
            int i = 0;
            for (Iterator it = col.iterator(); it.hasNext();) {
                BaseCluster cl = (BaseCluster) it.next();
                if (cl == null) {
                    logger.log(BasicLevel.ERROR, "null cl elt");
                } else {
                    sb[i++] = cl.getObjectName();
                }
            }
            return sb;
         */
    }

    /**
     * Handle Notifications from Discovery Service
     * @param event
     */
    public void discoveryNotification(DiscoveryEvent event) {
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "Get notification:" + event);
        }
        if (! event.getDomainName().equals(domainName)) {
            logger.log(BasicLevel.WARN, "Discovery notification from another domain. Forget it.");
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "Notification Domain = " + event.getDomainName());
            }
            return;
        }
        String state = event.getState();
        String svname = event.getServerName();

        // Look if the server is known
        ServerProxy proxy = findServerProxy(svname);
        if (proxy == null) {
            // Create an entry for this server
            proxy = new ServerProxy(this, svname);
            // give a name and register mbean
            ObjectName on;
            try {
                on = JonasObjectName.serverProxy(domainName, svname);
                proxy.setObjectName(on.toString());
            } catch (MalformedObjectNameException e) {
                logger.log(BasicLevel.ERROR, "MalformedObjectName - Cannot register MBEAN:" + svname);
            }
        }

        // If server is running, get its urls to establish a connection.
        if (state.equals(DiscoveryState.RUNNING)) {
            // If this is myself, just forget it
            if (svname.equals(masterName)) {
                logger.log(BasicLevel.DEBUG, "Yes, I'm already running, I knew that.");
                return;
            }
            // Establish the connection
            ArrayList urls = new ArrayList();
            String [] urlstr = event.getConnectorURL();
            for (int i = 0; i < urlstr.length; i++) {
                urls.add(urlstr[i]);
            }
            proxy.notifyStarting(urls);
            //retriev jonasBase, jonasRoot, javaHome to avoid retrieving on user request.
            proxy.retrieveClusterdInfos();
            notifyServerProxyRunning(proxy);
        } else if (state.equals(DiscoveryState.STOPPING)) {
            proxy.notifyStopping();
            // TODO Notify ClusterFactory's
        }
    }

    /**
     * Notify to all the cluster factories that a server started in the domain
     * @param sp the server's proxy
     */
    public void notifyServerProxyRunning(ServerProxy sp) {
        for (Iterator it = clusterFactoryList.iterator(); it.hasNext();) {
            ClusterFactory cf = (ClusterFactory) it.next();
            cf.notifyServer(sp);
        }
    }

    /**
     * Register a DeployAction
     * @return true if correctly aded in the list
     */
    public synchronized boolean registerDeployAction(DeployAction action) {
        // Don't add if already another similar action
        for (Iterator it = deployList.iterator(); it.hasNext(); ) {
            DeployAction act = (DeployAction) it.next();
            if (act.getFileName().equals(action.getFileName()) &&
                    act.getServerName().equals(action.getServerName())) {
                logger.log(BasicLevel.DEBUG, "Already a similar action");
                return false;
            }
        }
        return deployList.add(action);
    }

    /**
     * Get the list of server where a file is being deployed
     * @param filename file to deploy
     * @return server name
     */
    public synchronized String [] getDeployServers(String filename) {
        ArrayList slist = new ArrayList();
        for (Iterator it = deployList.iterator(); it.hasNext(); ) {
            DeployAction act = (DeployAction) it.next();
            if (act.getFileName().equals(filename)) {
                slist.add(act.getServerName());
            }
        }
        return (String []) slist.toArray(new String [0]);
    }

    /**
     * Get the current state (string form) of deployment operation
     * @param filename file to deploy
     * @param servername server where deployment is done
     * @return one of "progress","ok","fail"
     */
    public synchronized String getDeployState(String filename, String servername) {
        String ret = "error";
        for (Iterator it = deployList.iterator(); it.hasNext(); ) {
            DeployAction act = (DeployAction) it.next();
            if (act.getFileName().equals(filename) && act.getServerName().equals(servername)) {
                ret = act.getStateAsString();
                break;
            }
        }
        return ret;
    }

    /**
     * Get the error message associated to the error state
     * @param filename file to deploy
     * @param servername server where deployment is done
     * @return error message
     */
    public synchronized String getErrorMessage(String filename, String servername) {
        String ret = "";
        for (Iterator it = deployList.iterator(); it.hasNext(); ) {
            DeployAction act = (DeployAction) it.next();
            if (act.getFileName().equals(filename) && act.getServerName().equals(servername)) {
                ret = act.getErrorMessage();
            }
        }
        return ret;
    }

    /**
     * Forget all deploy information
     */
    public synchronized void forgetAllDeploy() {
        deployList.clear();
    }

    /**
     * Try to refresh the server states
     * TODO notify thread only.
     */
    public void refreshStates() {
        try {
            checkServerStates(false);
        } catch (JMException e) {
            logger.log(BasicLevel.DEBUG, "Cannot refresh states:", e);
        }
    }

    // ------------------------------------------------------------------------------
    // Thread associated to the DomainMonitor
    // ------------------------------------------------------------------------------

    /**
     * Check state of all remote servers in the domain
     * @throws JMException
     */
    private synchronized void checkServerStates(boolean readall) throws JMException {
        for (Iterator i = getServerList().iterator(); i.hasNext();) {
            ServerProxy proxy = (ServerProxy) i.next();
            proxy.checkit(readall);
        }
        for (Iterator i = getClusterDaemonList().iterator(); i.hasNext();) {
            ClusterDaemonProxy proxy = (ClusterDaemonProxy) i.next();
            proxy.checkit();
        }
    }

    /**
     * Monitoring thread that checks managed servers periodically
     */
    class StateMonitor extends Thread {

        /**
         * My DomainMonitor
         */
        private DomainMonitor domainMonitor;

        /**
         * Logger object to log events
         */
        private Logger logger = Log.getLogger("org.ow2.jonas.domain.management");

        /**
         * Default sampling period of 60 seconds
         */
        private long  samplingPeriod = 60000L;

        /**
         * To stop the thread
         */
        private boolean stopped = false;

        /**
         * Constructor
         * @param domainMonitor reference on the DomainMonitor instance
         * @param log the logger object
         */
        public StateMonitor(DomainMonitor domainMonitor) {
            super("DomainStateMonitor");
            setDaemon(true);
            this.domainMonitor = domainMonitor;
        }

        /**
         * Set the sampling period. Allows reconfiguration.
         * @param sec sampling period in sec.
         */
        public void setSamplingPeriod(int sec) {
            samplingPeriod = sec * 1000L;
        }

        /**
         * @return Sampling period in seconds
         * @return
         */
        public int getSamplingPeriod() {
            return new Long(samplingPeriod / 1000L).intValue();
        }

        /**
         * Stop this thread.
         */
        public void stopit() {
            stopped  = true;
        }

        /**
         * Start the state monitor thread
         */
        public void run() {
            while (!stopped) {
                try {
                    sleep(samplingPeriod);
                    domainMonitor.checkServerStates(true);
                } catch (JMException me) {
                    logger.log(BasicLevel.WARN, "Exception while checking server states", me);
                } catch (InterruptedException e) {
                    logger.log(BasicLevel.ERROR, "Interrupted", e);
                    break;
                }
            }
        }
    }

    /**
     * Return a JMX connection to a given server in the domain.
     * Use for that, the ServerProxy associated to the server.
     * @param serverName the name of the server
     * @return the connection provided by its ServerProxy if there is
     * a ServerProxy object in this domain that corresponds to the server
     * named serverName. Null if there is no such serverProxy or if the serverName
     * is null.
     */
    public MBeanServerConnection getConnection(String serverName) {

        // We want a connection to the local server
        if (serverName == null || jonasServerName.equals(serverName)) {
            return jmx.getJmxServerConnection();
        }

        // A connection to a remote server is possible only on the master
        if (! ismaster) {
             logger.log(BasicLevel.WARN, "This server is not the master");
             return null;
        }
        ServerProxy serverProxy = findServerProxy(serverName);
        if (serverProxy != null) {
            return serverProxy.getConnection();
        } else {
            return null;
        }
    }

    /**
     * Return the JMX ConnectorServerURLs of a given server in the domain.
     * Use for that, the ServerProxy associated to the server.
     * @param serverName the name of the server
     * @return the ConnectorServerURLs provided by its ServerProxy if there is
     * a ServerProxy object in this domain that corresponds to the server
     * named serverName. Null if thger is no such serverProxy or if the serverName
     * is null. Alse null if the ConnectorServerURLs list is null.
     */
    public String[] getConnectorServerURLs(String serverName) {
        if (serverName == null) {
            return null;
        }
        ServerProxy serverProxy = findServerProxy(serverName);
        if (serverProxy != null) {
            ArrayList urlList = null;
            urlList = serverProxy.getUrls();
            if (urlList == null) {
                return null;
            }
            String[] ret = new String[urlList.size()];
            Iterator it = urlList.iterator();
            for (int i = 0; i < urlList.size(); i++) {
                ret[i] = (String) it.next();
            }
            return ret;
        } else {
            return null;
        }
    }

    /**
     * @param started if true, return only the servers which are in RUNNING state
     * anyway, don't pat attention to the servers state
     * @return an array containing the OBJECT_NAMEs of the J2EEServer MBeans
     * associated to the managed servers in this domain.
     */
    public String[] getServers(boolean started) {
        Collection col = getServerList(); // not null, maybe empty
        ArrayList al = new ArrayList();
        for (Iterator it = col.iterator(); it.hasNext();) {
            ServerProxy proxy = (ServerProxy) it.next();
            boolean adit = true;
            if (started) {
                if (proxy.getServerState() != ServerProxy.RUNNING) {
                    adit = false;
                }
            }
            if (adit) {
                al.add(proxy.getJ2eeObjectName());
            }
        }
        return (String []) al.toArray(new String [al.size()]);
    }

    /**
     * @return an array containing the names of the managed servers in this domain.
     */
    public String[] getServerNames() {
        Collection col = getServerList();
        String[] names = new String[col.size()];
        int i = 0;
        for (Iterator it = col.iterator(); it.hasNext();) {
            ServerProxy proxy = (ServerProxy) it.next();
            names[i++] = proxy.getName();
        }
        return names;
    }

    /**
     * @param clusterName the cluster name
     * @return an array containing the names of the managed servers in the cluster.
     */
    public String[] getServerNames(String clusterName) {
        String[] names = null;
        BaseCluster cl = findCluster(clusterName);
        if (cl != null) {
            int nbServers = cl.getNbMembers();
            names = new String[nbServers];
            if (nbServers == 0) {
                return names;
            }
            Iterator sps = cl.getServerProxyList().iterator();
            int i = 0;
            while (sps.hasNext()) {
                ServerProxy sp = (ServerProxy) sps.next();
                names[i++] = sp.getServerName();
            }
        }
        return names;
    }

    /**
     * Return the state of a server in the domain. Used by JonasAdmin.
     * @param serverName the server name
     * @return Get the state of a server in the domain
     */
    public String getServerState(String serverName) {
        ServerProxy proxy = findServerProxy(serverName);
        if (proxy != null) {
            return proxy.getState();
        }
        return null;
    }

    /**
     * Return the state of a cluster in the domain. Used by JonasAdmin.
     * @param clusterName the cluster name
     * @return Get the state of a cluster in the domain
     */
    public String getClusterState(String clusterName) {
        BaseCluster cl = findCluster(clusterName);
        if (cl != null) {
            return cl.getState();
        }
        return null;
    }

    /**
     * Return the state of a cluster daemon in the domain. Used by JonasAdmin.
     * @param clusterdaemonName the cluster daemon name
     * @return Get the state of a cluster daemon in the domain
     */
    public String getClusterdaemonState(String clusterdaemonName) {
        ClusterDaemonProxy cdProxy = findClusterDaemonProxy(clusterdaemonName);
        if (cdProxy != null) {
            return cdProxy.getState();
        }
        return null;
    }

    /**
     * Return the type of a cluster in the domain. Used by JonasAdmin.
     * @param clusterName the cluster name
     * @return Get the type of a cluster in the domain
     */
    public String getClusterType(String clusterName) {
        BaseCluster cl = findCluster(clusterName);
        if (cl != null) {
            return cl.getType();
        }
        return null;
    }
    /**
     * @return an array containing the OBJECT_NAMEs of the ClusterDaemons in this domain
     */
    @SuppressWarnings("unchecked")
    public String[] getClusterDaemons() {
        String[] sb = null;
        Collection col = getClusterDaemonList();
        sb = new String[col.size()];
        int i = 0;
        for (Iterator it = col.iterator(); it.hasNext();) {
            ClusterDaemonProxy cdp = (ClusterDaemonProxy) it.next();
            sb[i++] = cdp.getObjectName();
        }
        return sb;
    }

    /**
     * Create a logical cluster
     * @param name the name of the cluster
     * @return the OBJECT_NAME of the associated MBean
     */
    public String createCluster(String name) {
        BaseCluster cl = createLogicalCluster(name);
        return cl.getObjectName();
    }
    /**
     * Start a managed JOnAS Server
     * @param serverName name of the server
     */
    public void startServer(String serverName) {
        ServerProxy proxy = findServerProxy(serverName);
        proxy.startit();
    }

    /**
     * Stop a managed JOnAS Server
     * @param serverName name of the server
     */
    public void stopServer(String serverName) {
        ServerProxy proxy = findServerProxy(serverName);
        proxy.stopit();
    }

    /**
     * Deploy a module on a target which may be a server or a cluster
     * @param target target name
     * @param fileName file containing the module
     */
    public void deployOnTarget(String target, String fileName) {
        ServerProxy sp = findServerProxy(target);
        if (sp != null) {
            // The target is a known server in the domain
            sp.deployModule(fileName);
        } else {
            // maybe the target is a cluster
            BaseCluster cl = findCluster(target);
            if (cl != null) {
                cl.deployModule(fileName);
            } else {
                logger.log(BasicLevel.ERROR, "Can't deploy on unknown target " + target);
            }
        }
    }

    /**
     * Upload a module on a target which may be a server or a cluster
     * @param target target name
     * @param fileName file containing the module
     * @param replaceExisting true if the file can replace an existing one (same name)
     */
    public void uploadDeployOnTarget(String target, String fileName, boolean replaceExisting) {
        ServerProxy sp = findServerProxy(target);
        if (sp != null) {
            // The target is a known server in the domain
            sp.uploadDeployModule(fileName, replaceExisting);
        } else {
            // maybe the target is a cluster
            BaseCluster cl = findCluster(target);
            if (cl != null) {
                cl.uploadDeployModule(fileName, replaceExisting);
            } else {
                logger.log(BasicLevel.ERROR, "Can't get deploy on unknown target " + target);
            }
        }
    }
    /**
     * Undeploy a module on a target which may be a server or a cluster
     * @param target target name
     * @param fileName file containing the module
     */
    public void unDeployOnTarget(String target, String fileName) {
        ServerProxy sp = findServerProxy(target);
        if (sp != null) {
            // The target is a known server in the domain
            sp.undeployModule(fileName);
        } else {
            // maybe the target is a cluster
            BaseCluster cl = findCluster(target);
            if (cl != null) {
                cl.undeployModule(fileName);
            } else {
                logger.log(BasicLevel.ERROR, "Can't get undeploy on unknown target " + target);
            }
        }
    }

    /**
     * Stop a server or a cluster
     * @param target the name of the target to be stoped
     * @return true is success, false otherwise
     * @throws JMException problem when trying to stop
     */
    public boolean stopRemoteTarget(String target) throws JMException {
        BaseCluster cluster = findCluster(target);
        if (cluster != null) {
            cluster.stopit();
            return true;
        }
        ServerProxy server = findServerProxy(target);
        if (server != null) {
            // this target is a server
            server.stopit();
            return true;
        }
        return false;
    }

    /**
     * Start a server or a cluster
     * @param target the name of the target to be started
     * @return true is success, false otherwise
     * @throws JMException problem when trying to start
     */
    public boolean startRemoteTarget(String target) throws JMException {
        BaseCluster cluster = findCluster(target);
        if (cluster != null) {
            cluster.startit();
            return true;
        }
        ServerProxy server = findServerProxy(target);
        if (server != null) {
            // this target is a server
            server.startit();
            return true;
        }
        return false;
    }

    /**
     * Upload and deploy a module on a target.
     * @param target the target name
     * @param filename the file containing the module
     * @return true is success, false otherwise
     */
    public boolean uploadDeployFileOn(String target, String filename) {
        BaseCluster cluster = findCluster(target);
        if (cluster != null) {
            // this target is a cluster
            cluster.uploadDeployModule(filename, true);
            return true;
        }
        ServerProxy server = findServerProxy(target);
        if (server != null) {
            // this target is a server
            server.uploadDeployModule(filename, true);
            return true;
        }
        return false;
    }

    /**
     * Check if a cluster exists.
     * @param clusterName the name of the cluster to check
     * @return true if a cluster with this name exists
     */
    public boolean isCluster(String clusterName) {
        BaseCluster cluster = findCluster(clusterName);
        if (cluster != null) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * @param clusterName Name of the cluster
     * @return the servers in the domain that doesn't belong to the cluster
     */
    public String[] getServersNotInCluster(String clusterName) {
        String[] allServers = getServerNames();
        BaseCluster clust = findCluster(clusterName);
        if (clust != null) {
            String[] allServersInCluster = getServerNames(clusterName);
            if (allServersInCluster.length == 0) {
                return allServers;
            } else {
                int nbServers = allServers.length - allServersInCluster.length;
                String[] allServersNotInCluster = new String[nbServers];
                int j = 0;
                for (int i = 0; i < allServers.length; i++) {
                    String server = allServers[i];
                    if (!clust.isMember(server)) {
                        try {
                            allServersNotInCluster[j++] = server;
                        } catch (IndexOutOfBoundsException be) {
                            logger.log(BasicLevel.DEBUG, "Can't get servers not in cluster " + clusterName + " correctly");
                            break;
                        }
                    }
                }
                return allServersNotInCluster;
            }
        } else {
            return allServers;
        }
    }

    /**
     * @param clusterName Name of the cluster
     * @return the servers in the domain that belong to the cluster
     */
    public String[] getServersInCluster(String clusterName) {
        return getServerNames(clusterName);
    }

    /**
     * Return the cluster daemon name of a server in the domain. Used by JonasAdmin.
     * @param serverName the server name
     * @return Get the cluster daemon name of a server in the domain
     */
    public String getServerClusterdaemon(String serverName) {
        ServerProxy proxy = findServerProxy(serverName);
        if (proxy != null) {
            return proxy.getClusterDaemonName();
        }
        return null;
    }
}