/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999-2007 Bull S.A.S.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.server;

import java.rmi.RemoteException;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Properties;
import java.util.Vector;

import javax.management.JMException;
import javax.naming.InitialContext;
import javax.naming.NameClassPair;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.rmi.PortableRemoteObject;

import org.ow2.jonas.adm.AdmException;
import org.ow2.jonas.adm.AdmInterface;
import org.ow2.jonas.dbm.internal.JOnASDataBaseManagerService;
import org.ow2.jonas.dbm.internal.cm.ConnectionManager;
import org.ow2.jonas.dbm.internal.cm.Pool;
import org.ow2.jonas.ear.EarService;
import org.ow2.jonas.ear.EarServiceException;
import org.ow2.jonas.ejb.internal.JOnASEJBService;
import org.ow2.jonas.lib.bootstrap.JProp;
import org.ow2.jonas.lib.ejb21.Container;
import org.ow2.jonas.lib.ejb21.JFactory;
import org.ow2.jonas.lib.log.LogManagement;
import org.ow2.jonas.lib.management.domain.J2EEDomain;
import org.ow2.jonas.lib.naming.SingletonNamingManager;
import org.ow2.jonas.resource.ResourceService;
import org.ow2.jonas.service.ServiceException;
import org.ow2.jonas.service.manager.ServiceManager;
import org.ow2.jonas.tm.jotm.JOTMTransactionService;
import org.ow2.jonas.web.JWebContainerService;


/**
 * This class implements a remote interface used for administering the server.
 * @author Philippe Coq
 * @author Philippe Durieux
 * @author Marc Dutoo, JOnAS 2.4
 * @author Murad Meghani (Murad.Meghani@compuware.com) killServer and stopServer
 * @author Florent Benoit & Ludovic Bert : Methods for wars and ear files JOnAS
 *         2.5 2002.06
 * @author Dean Jennings : add deployFile
 */
public class Adm extends PortableRemoteObject implements AdmInterface {

    /**
     * ServiceManager instance.
     */
    private ServiceManager sm = null;

    /**
     * EJBService of this JOnAS instance.
     */
    private JOnASEJBService ejbserv = null;

    /**
     * DBMService of this JOnAS instance.
     */
    private JOnASDataBaseManagerService dbmserv = null;

    /**
     * TransactionService of this JOnAS instance.
     */
    private JOTMTransactionService tm = null;

    /**
     * JOnAS configuration.
     */
    private JProp jonasProperties = null;

    private InitialContext ictx = null;

    /**
     * Does JOnAS have the EJBService deployed ?
     */
    private boolean isEJBContainer;

    /**
     * Initial state : NotReady
     */
    private int serverState = NOT_READY;

    /**
     * Sleep before exit of JVM
     */
    public static final int SLEEP_DELAY = 2000;

    /**
     * Reference to the web container service
     */
    private JWebContainerService webContainerService = null;

    /**
     * Reference to the ear service
     */
    private EarService earService = null;

    /**
     * Reference to the resource service
     */
    private ResourceService resourceService = null;

    /**
     * Reference on the J2EEDomain
     * Used to access other servers and clusters in the domain.
     */
    J2EEDomain mydomain = null;

    /**
     * Adm constructor
     * @throws RemoteException if problem
     */
    public Adm(JProp jp) throws RemoteException, NamingException, ServiceException, Exception {

        jonasProperties = jp;

        // Get Service references
        sm = ServiceManager.getInstance();

        // Get an Initial Context and
        // registers the JOnAS Server Name to JNDI
        String name = null;
        name = jonasProperties.getValue("jonas.name", "jonas") + ADMNAME_SUFFIX;
        ictx = SingletonNamingManager.getInstance().getInitialContext();
        ictx.rebind(name, this);

        // Keep a reference on the J2EEDomain
        mydomain = Server.getInstance().getJ2EEDomain();

    }

    private JOTMTransactionService getTM() {
        if (tm == null) {
            try {
                tm = (JOTMTransactionService) sm.getTransactionService();
            } catch (ServiceException e) {
            }
        }
        return tm;
    }

    private JOnASDataBaseManagerService getDBM() {
        if (dbmserv == null) {
            try {
                dbmserv = (JOnASDataBaseManagerService) sm.getDataBaseService();
            }catch (ServiceException e) {
            }
        }
        return dbmserv;
    }

    /**
     * Get the web container service We can use methods provided by the
     * interface for MBean
     */
    private JWebContainerService getWebContainerService() {
        if (webContainerService == null) {
            try {
                webContainerService = (JWebContainerService) sm.getWebContainerService();
            } catch (ServiceException e) {
                // not started
            }
        }
        return webContainerService;
    }

    /**
     * Get the ear service.
     */
    private EarService getEarService() {
        if (earService == null) {
            try {
                earService = (EarService) sm.getEarService();
            } catch (Exception e) {
                // not started
            }
        }
        return earService;
    }

    /**
     * Get the resource service We can use methods provided by the interface for
     * MBean
     */
    private ResourceService getResourceService() {
        if (resourceService == null) {
            try {
                resourceService = (ResourceService) sm.getResourceService();
            } catch (Exception e) {
                // not started
            }
        }
        return resourceService;
    }

    // ------------------------------------------------------------------
    // AdmInterface implementation
    // ------------------------------------------------------------------

    /**
     * get Topics. Assumes that all Loggers are TopicalLoggers.
     */
    public String[] getTopics() throws RemoteException {
        return LogManagement.getInstance().getTopics();
    }

    /**
     * get Topic Level
     */
    public String getTopicLevel(String topic) throws RemoteException {
        return LogManagement.getInstance().getTopicLevel(topic);
    }

    /**
     * set Topic Level
     */
    public void setTopicLevel(String topic, String l) throws RemoteException {
        LogManagement.getInstance().setTopicLevel(topic, l);
    }

    /**
     * Create a container and load beans in it
     * @param fileName name of the ejb-jar or xml file
     */
    public void addBeans(String fileName) throws RemoteException {
        if (!isEJBContainer) {
            return;
        }
        try {
            ejbserv.createContainer(fileName);
        } catch (Exception e) {
            throw new RemoteException("Cannot add beans", e);
        }
    }

    /**
     * Deploy a given ear file with the help of the ear service.
     * @param fileName the name of the ear file.
     * @throws RemoteException if rmi call failed.
     * @throws EarServiceException if the deployment failed.
     */
    public void addEar(String fileName) throws RemoteException, AdmException {
        if (getEarService() != null) {
            try {
                getEarService().deployEar(fileName);
            } catch (AdmException ear) {
                throw ear;
            } catch (Exception e) {
                throw new RemoteException("Unable to deploy the EAR '" + fileName + "'.", e);
            }
        } else {
            throw new AdmException("Can't add the Ear file " + fileName + ". The ear service is not started");
        }
    }

    /**
     * Test if the specified filename is already deployed or not
     * @param fileName the name of the ear file.
     * @return true if the ear is deployed, else false.
     * @throws RemoteException if rmi call failed.
     */
    public boolean isEarLoaded(String fileName) throws RemoteException, AdmException {
        boolean isLoaded = false;
        if (getEarService() != null) {
            isLoaded = getEarService().isEarDeployed(fileName);
        } else {
            throw new AdmException("Can't test if the ear file " + fileName
                    + " is deployed or not. The ear service is not started");
        }
        return isLoaded;
    }

    /**
     * Deploy a given rar file with the help of the resource service.
     * @param fileName the name of the rar file.
     * @throws RemoteException if rmi call failed.
     * @throws AdmException if the deployment failed.
     */
    public void addRar(String fileName) throws RemoteException, AdmException {
        if (getResourceService() != null) {
            try {
                getResourceService().deployRar(fileName);
            } catch (Exception e) {
                throw new RemoteException("Unable to deploy the RAR '" + fileName + "'.", e);
            }
        } else {
            throw new AdmException("Can't add the Rar file " + fileName
                    + ". The resource service is not started");
        }
    }

    /**
     * Test if the specified filename is already deployed or not
     * @param fileName the name of the rar file.
     * @return true if the rar is deployed, else false.
     * @throws RemoteException if rmi call failed.
     * @throws AdmException if unable to get resource service
     */
    public boolean isRarLoaded(String fileName) throws RemoteException, AdmException {
        boolean isLoaded = false;
        if (getResourceService() != null) {
            isLoaded = getResourceService().isRarLoaded(fileName);
        } else {
            throw new AdmException("Can't test if the rar file " + fileName
                    + " is deployed or not. The rar service is not started");
        }
        return isLoaded;
    }

    /**
     * Deploy a given war file with the help of the web container service.
     * @param fileName the name of the war file.
     * @throws RemoteException if rmi call failed.
     * @throws AdmException if the deployment failed.
     */
    public void addWar(String fileName) throws RemoteException, AdmException {
        if (getWebContainerService() != null) {
            getWebContainerService().registerWar(fileName);
        } else {
            throw new AdmException("Can't add the war file " + fileName
                    + ". The web container service is not started");
        }
    }

    /**
     * Test if the specified filename is already deployed or not
     * @param fileName the name of the war file.
     * @return true if the war is deployed, else false.
     * @throws RemoteException if rmi call failed.
     */
    public boolean isWarLoaded(String fileName) throws RemoteException, AdmException {
        boolean isLoaded = false;
        if (getWebContainerService() != null) {
            isLoaded = getWebContainerService().isWarLoaded(fileName);
        } else {
            throw new AdmException("Can't test if the war file " + fileName
                    + " is deployed or not. The war service is not started");
        }
        return isLoaded;
    }

    /**
     * UnDeploy a given ear file with the help of the ear service.
     * @param fileName the name of the ear file.
     * @throws RemoteException if rmi call failed.
     * @throws AdmException if the undeployment failed.
     */
    public void removeEar(String fileName) throws RemoteException, AdmException {
        if (getEarService() != null) {
            try {
                getEarService().unDeployEar(fileName);
                // needed here to run also the distributed garbage collector
                runGC();
            } catch (AdmException ear) {
                throw ear;
            } catch (Exception e) {
                throw new RemoteException("Cannot remove ear", e);
            }
        } else {
            throw new AdmException("Can't remove of the Ear file " + fileName
                    + ". The ear service is not started");
        }
    }

    /**
     * UnDeploy a given rar file with the help of the resource service.
     * @param fileName the name of the rar file.
     * @throws RemoteException if rmi call failed.
     * @throws AdmException if the undeployment failed.
     */
    public void removeRar(String fileName) throws RemoteException, AdmException {
        if (getResourceService() != null) {
            try {
                getResourceService().unDeployRar(fileName);
            } catch (Exception e) {
                throw new RemoteException("Unable to undeploy the RAR '" + fileName + "'.", e);
            }
            // needed here to run also the distributed garbage collector
            runGC();
        } else {
            throw new AdmException("Can't remove of the Rar file " + fileName
                    + ". The resource service is not started");
        }
    }

    /**
     * UnDeploy a given war file with the help of the web container service.
     * @param fileName the name of the war file.
     * @throws RemoteException if rmi call failed.
     * @throws AdmException if the undeployment failed.
     */
    public void removeWar(String fileName) throws RemoteException, AdmException {
        if (getWebContainerService() != null) {
            getWebContainerService().unRegisterWar(fileName);
            // needed here to run also the distributed garbage collector
            runGC();
        } else {
            throw new AdmException("Can't remove of the war file " + fileName
                    + ". The web container service is not started");
        }
    }

    /**
     * Remove the container identified by fileName and remove all beans in it
     * @param fileName name of the ejb-jar or xml file
     */
    public void removeBeans(String fileName) throws RemoteException {
        if (!isEJBContainer) {
            return;
        }
        try {
            ejbserv.removeContainer(fileName);
        } catch (Exception e) {
            throw new RemoteException("Cannot remove bean", e);
        }
        // needed here to run also the distributed garbage collector
        runGC();
    }

    /**
     * returns true if beans are already loaded in server.
     * @param fileName name of the ejb-jar or xml file
     */
    public boolean isLoaded(String fileName) throws RemoteException {
        return (ejbserv.isJarLoaded(fileName));
    }

    public String dumpCustom() throws RemoteException {
        String ret = "";
        if (!isEJBContainer) {
            return ret;
        }
        // JVM
        ret += Server.jvmInfos();

        // Transaction Service
        if (getTM() != null) {
            ret += "TM timeout=" + getTM().getTimeout();
            ret += "\n";
        }

        // datasources
        if (getDBM() != null) {
            Collection dslist = getDBM().getDSList();
            for (Iterator i = dslist.iterator(); i.hasNext(); ) {
                ConnectionManager cm = (ConnectionManager) i.next();
                ret += cm.getDSName() + ":lockPolicy=" + cm.getTransactionIsolation();
                Pool pool = cm.getPool();
                ret += ":minPoolSize=" + pool.getPoolMin();
                ret += ":maxPoolSize=" + pool.getPoolMax();
                ret += ":maxOpenTime=" + pool.getMaxOpenTime();
                ret += ":maxWaitTime=" + pool.getMaxWaitTime();
                ret += ":maxWaiters=" + pool.getMaxWaiters();
                ret += ":pstmtMax=" + cm.getPstmtMax();
                ret += "\n";
            }
        }

        // EJB Containers
        Container[] lcont = (Container[]) ejbserv.listContainers();
        for (int j = 0; j < lcont.length; j++) {
            Container cont = lcont[j];
            String contname = cont.getName();
            String[] lbn = cont.listBeanNames();
            for (int k = 0; k < lbn.length; k++) {
                JFactory bf = (JFactory) cont.getBeanFactory(lbn[k]);
                ret += contname + ":=" + lbn[k];
                ret += ":minPoolSize=" + bf.getMinPoolSize();
                ret += ":maxCacheSize=" + bf.getMaxCacheSize();
                ret += "\n";
            }
        }
        return ret;
    }

    /**
     * List beans of all JOnAS containers
     */
    public String[] listBeans() throws RemoteException {

        if (!isEJBContainer) {
            return new String[0];
        }

        Container[] lcont = ejbserv.listContainers();
        Vector lbeans = new Vector();
        for (int j = 0; j < lcont.length; j++) {
            String[] lbn = lcont[j].listBeanNames();
            String contname = lcont[j].getName();
            for (int k = 0; k < lbn.length; k++) {
                String s = new String(contname + ": " + lbn[k]);
                lbeans.addElement(s);
            }
        }
        String[] ret = new String[lbeans.size()];
        lbeans.copyInto(ret);

        return ret;
    }

    /**
     * List JNDI context
     */
    public Vector listContext() throws RemoteException {
        String name = null;
        Vector ret = new Vector();
        try {
            NamingEnumeration ne;
            ne = SingletonNamingManager.getInstance().getInitialContext().list("");
            for (Enumeration e = ne; e.hasMoreElements();) {
                name = ((NameClassPair) e.nextElement()).getName();
                ret.addElement(name);
            }
        } catch (Exception ex) {
            throw new RemoteException("Cannot list JNDI context", ex);
        }
        return ret;
    }

    /**
     * List Environment (configuration properties provided by the configuration
     * file).
     */
    public Properties listEnv() {
        return jonasProperties.getConfigFileEnv();
    }

    /**
     * Stop the Server without stopping the JVM
     */
    public void stopServer() throws RemoteException {
        try {
            serverState = STOPPED;
            // keep registry alive a little while (for jonas admin) in case of
            // error at launch
            Thread.sleep(5000);
            sm.stopServices();
        } catch (ServiceException e) {
            throw new RemoteException("Cannot stop services: ", e);
        } catch (InterruptedException e) {
        }
    }

    /**
     * Stop the Server and stop the JVM
     */
    public void killServer() throws RemoteException {
        // Stops the JOnAS's services
        stopServer();

        // Unregisters the JOnAS Server Name in the JNDI if the registry is still up (should
        // be the case whether it is shared.
        String name = null;
        try {
        name = jonasProperties.getValue("jonas.name", "jonas") + ADMNAME_SUFFIX;
            ictx = SingletonNamingManager.getInstance().getInitialContext();
            ictx.unbind(name);
        } catch (Exception e) {
            //e.printStackTrace();
        }

        // Delay the exit of the JVM so the link client/server is not broken.
        // client call will return before the end of the JVM on server side.
        new Thread(new Runnable() {

            public void run() {
                try {
                    // Wait before exit
                    Thread.sleep(SLEEP_DELAY);
                } catch (InterruptedException ie) {
                    ie.printStackTrace();
                    throw new IllegalStateException("Cannot wait: " + ie.getMessage());
                }
                System.exit(0);
            }
        }).start();

    }

    /**
     * To test if the server is ready
     * @return int 0=not ready, 1=ready, 2=stopped
     */
    public int getServerState() throws RemoteException {
        return serverState;
    }

    /**
     * To test if the server is an EJB container
     */
    public boolean isEJBContainer() throws RemoteException {
        return isEJBContainer;
    }

    /**
     * set the default value for transaction timeout
     */
    public void setTransactionTimeout(int timeout) throws RemoteException {
        getTM().setTimeout(timeout);
    }

    /**
     * run the garbage collector
     */
    public void runGC() throws RemoteException {
        Runtime.getRuntime().gc();
    }

    /**
     * sync all entity instances outside transactions
     * @param passivate passivate instances after synchronization.
     */
    public void syncAllEntities(boolean passivate) throws RemoteException {
        if (!isEJBContainer) {
            return;
        }
        ejbserv.syncAllEntities(passivate);
    }

    // ------------------------------------------------------------------
    // other public methods
    // ------------------------------------------------------------------

    /**
     * server is ready
     */
    public void serverReady(boolean isEJB) {
        serverState = READY;
        isEJBContainer = isEJB;
        if (isEJBContainer) {
            ejbserv = (JOnASEJBService) sm.getEjbService();
        }
    }

    /**
     * Start remote servers or clusters
     * @param target List of clusters or servers to start
     * @throws RemoteException
     */
    public void startRemoteServers(String [] target) throws RemoteException {
        if (!mydomain.isMaster()) {
            throw new RemoteException("Can't start remote targets as this server is not a master");
        }
        for (int i = 0; i < target.length; i++) {
            String mytarget = target[i];
            try {
                boolean succes = mydomain.startRemoteTarget(mytarget);
                if (!succes) {
                    throw new RemoteException("unknown target: " + mytarget);
                }
            } catch (JMException e) {
                throw new RemoteException("Can't start target " + mytarget + e.toString());
            }
        }
    }

    /**
     * Stop remote servers or clusters
     * @param target List of clusters or servers to stop
     * @throws RemoteException
     */
    public void stopRemoteServers(String [] target) throws RemoteException {
        if (!mydomain.isMaster()) {
            throw new RemoteException("Can't stop remote targets as this server is not a master");
        }
        for (int i = 0; i < target.length; i++) {
            String mytarget = target[i];
            try {
                boolean succes = mydomain.stopRemoteTarget(mytarget);
                if (!succes) {
                    throw new RemoteException("unknown target: " + mytarget);
                }
            } catch (JMException e) {
                throw new RemoteException("Can't stop target " + mytarget + e.toString());
            }
        }
    }

    /**
     * Deploy file on another server or a cluster.
     * @param filename Name of the file
     * @param target List of clusters or servers where to deploy the file.
     * @throws RemoteException
     */
    public void deployFileOn(String filename, String [] target) throws RemoteException {
        if (!mydomain.isMaster()) {
            throw new RemoteException("Can't stop remote targets as this server is not a master");
        }
        for (int i = 0; i < target.length; i++) {
            String mytarget = target[i];
            boolean succes = mydomain.uploadDeployFileOn(mytarget, filename);
            if (!succes) {
                throw new RemoteException("unknown target: " + mytarget);
            }
        }
    }

}
