/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999-2005 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */
package org.ow2.jonas.server;

// Java imports
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.rmi.RemoteException;
import java.util.ArrayList;

import javax.enterprise.deploy.shared.ModuleType;
import javax.management.MBeanException;
import javax.management.NotificationEmitter;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;

import org.apache.commons.modeler.BaseModelMBean;
import org.ow2.jonas.deployer.DeployerManagerService;
import org.ow2.jonas.deployment.web.WebContainerDeploymentDescException;
import org.ow2.jonas.deployment.web.wrapper.WebManagerWrapper;
import org.ow2.jonas.ear.EarServiceException;
import org.ow2.jonas.ear.internal.JOnASEARServiceMBean;
import org.ow2.jonas.ejb.internal.JOnASEJBServiceMBean;
import org.ow2.jonas.generators.genclientstub.wrapper.ClientGenStubWrapper;
import org.ow2.jonas.generators.wsgen.wrapper.WsGenWrapper;
import org.ow2.jonas.lib.bootstrap.JProp;
import org.ow2.jonas.resource.ResourceServiceException;
import org.ow2.jonas.resource.internal.JOnASResourceServiceMBean;
import org.ow2.jonas.service.manager.ServiceManager;
import org.ow2.jonas.web.JWebContainerServiceException;
import org.ow2.jonas.web.base.BaseWebContainerServiceMBean;
import org.ow2.util.ee.deploy.api.archive.IArchive;
import org.ow2.util.ee.deploy.api.deployable.EARDeployable;
import org.ow2.util.ee.deploy.api.deployable.EJBDeployable;
import org.ow2.util.ee.deploy.api.deployable.IDeployable;
import org.ow2.util.ee.deploy.api.deployable.RARDeployable;
import org.ow2.util.ee.deploy.api.deployer.DeployerException;
import org.ow2.util.ee.deploy.impl.archive.ArchiveManager;
import org.ow2.util.ee.deploy.impl.helper.DeployableHelper;
import org.ow2.util.ee.deploy.impl.helper.DeployableHelperException;
import org.ow2.util.file.FileUtils;
import org.ow2.util.file.FileUtilsException;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * MBean class for Server management. Two classes are used to provide all
 * management services : J2EEServer and J2EEServerMBean. This class provide the
 * notification's process when add or remove resources.
 * @author Adriana Danes
 * @author Michel-Ange Anton
 */

public class J2EEServerMBean extends BaseModelMBean {

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(J2EEServerMBean.class);

    /**
     * Folder to create in tmp folder.
     */
    private static final String DEFAULT_FOLDER = "JOnAS-Deployer";


    /**
     * Default constructor.
     * @throws MBeanException if super constructor fails
     */
    public J2EEServerMBean() throws MBeanException {
        super();
    }

    // --------------------Public methods implementing exposed interface

    /**
     * All The MBean names corresponding to the deployed J2EEModules (ear, jar,
     * war, rar).
     * @return The String's array of deployed modules objects names
     */
    public String[] getDeployedObjects() {
        ArrayList al = (ArrayList) ((J2EEServer) (this.resource)).getDeployedObjects();
        return (String[]) al.toArray(new String[al.size()]);
    }

    /**
     * All The MBean names corresponding to the deployed J2EEResources.
     * @return The String's array of deployed resources objects names
     */
    public String[] getResources() {
        ArrayList al = (ArrayList) ((J2EEServer) (this.resource)).getResources();
        return (String[]) al.toArray(new String[al.size()]);
    }

    /**
     * All The MBean names corresponding to the used JVMs.
     * @return The String's array of used JVMs objects names
     */
    public String[] getJavaVMs() {
        ArrayList al = (ArrayList) ((J2EEServer) (this.resource)).getJavaVMs();
        return (String[]) al.toArray(new String[al.size()]);
    }

    /**
     * Add a new listener.
     * @param pListner Listener to notify
     * @param pFilter Notification filter
     * @param pHandback ??
     * @throws java.lang.IllegalArgumentException if notification is not done
     */
    public void addNotificationListener(NotificationListener pListner, NotificationFilter pFilter,
            java.lang.Object pHandback) throws java.lang.IllegalArgumentException {
        ((NotificationEmitter) (this.resource)).addNotificationListener(pListner, pFilter, pHandback);
    }

    /**
     * Apply WsGen on a  local file and generate webservices artifacts
     * @param pathname : local path name to the component
     * @param unpacked true if we want to have an unpacked directory as result
     * @return Returns the path to the modified archive
     */
    public String wsgenLocalFile(String pathname, Boolean unpacked) {
        try {
               WsGenWrapper wsgen = new WsGenWrapper();
               String outputFilename = wsgen.callWsGenExecute(pathname, unpacked);
               boolean modified = wsgen.callWsGenIsInputModifed();
               if (modified) {
                   return outputFilename;
               } else {
                   return null;
               }
        } catch (Exception e) {
            logger.warn("Cannot generate web services for this component : '" + pathname + ".", e);
            return null;
        }
    }

    /**
     * Dump the given bytes to a local file and then deploy this file to a local deployer.
     * @param fileName the name of the file to deploy
     * @param fileContent the content of the given file
     */
    public void deployFile(final String fileName, final byte[] fileContent) {
        // Deploy this local file
        deploy(dumpFile(fileName, fileContent));
    }

    /**
     * Gets a deployable for a given file.
     * @param fileName the name of the file
     * @return a deployable for the given filename
     */
    protected IDeployable getDeployable(final String fileName) {
        // Get File
        File file = new File(fileName);

        // check file
        if (!file.exists()) {
            throw new RuntimeException("The file '" + fileName + "' is not present on the filesystem.");
        }

        // Else, get the deployable
        IArchive archive = ArchiveManager.getInstance().getArchive(file);
        if (archive == null) {
            logger.error("No archive found for the invalid file ''{0}''", file);
            throw new RuntimeException("No archive found for the invalid file '" + file + "'.");
        }
        IDeployable deployable;
        try {
            deployable = DeployableHelper.getDeployable(archive);
        } catch (DeployableHelperException e) {
            logger.error("Cannot get a deployable for the archive ''{0}''", archive, e);
            throw new RuntimeException("Cannot get a deployable for the archive '" + archive + "' : " + e.getMessage());
        }

        return deployable;
    }


    /**
     * Deploy a file to a local deployer.
     * @param fileName the name of the file to deploy
     */
    public void deploy(final String fileName) {

        logger.info("Deploying ''{0}''", fileName);

        // check if the given entry is a valid file
        File f = new File(fileName);
        if (!f.isAbsolute()) {
            throw new RuntimeException("The given filename '" + fileName + "' is not an absolute file.");
        }

        IDeployable deployable = getDeployable(fileName);

        try {
            DeployerManagerService.getDeployerManager().deploy(deployable);
        } catch (DeployerException e) {
            logger.error("Cannot deploy the deployable ''{0}''", deployable, e);
            throw new RuntimeException("Cannot deploy the deployable '" + deployable + "' : " + e.getMessage());
        }
    }

    /**
     * Undeploy a file from a local deployer.
     * @param fileName the name of the file to undeploy
     */
    public void undeploy(final String fileName) {

        logger.info("Undeploying ''{0}''", fileName);

        IDeployable deployable = getDeployable(fileName);

        try {
            DeployerManagerService.getDeployerManager().undeploy(deployable);
        } catch (DeployerException e) {
            logger.error("Cannot undeploy the deployable ''{0}''", deployable, e);
            throw new RuntimeException("Cannot undeploy the deployable '" + deployable + "' : " + e.getMessage());
        }
    }
    /**
     * Dump the given bytes to a local file and then return the path to this
     * file.
     * @param fileName the name of the file to deploy
     * @param fileContent the content of the given file
     * @return the path of the deployed file
     */
    public String dumpFile(final String fileName, final byte[] fileContent) {
        logger.info("Dump file to the local filesystem with the name = ''{0}''.", fileName);

        // Dump the file on a temporary file
        File rootFolder = new File(System.getProperty("java.io.tmpdir") + File.separator + DEFAULT_FOLDER);
        rootFolder.mkdirs();

        // Create file in this folder
        File file =  new File(rootFolder, fileName);

        // Dump the content on this file
        FileOutputStream out = null;
        try {
            out = new FileOutputStream(file);
        } catch (FileNotFoundException e) {
            throw new IllegalStateException("Cannot build an outputstream on file '" + file + "'.", e);
        }

        // Write the array of bytes
        try {
            out.write(fileContent);
        } catch (IOException e) {
            throw new IllegalStateException("Cannot write byte in outputstream", e);
        }

        // Flush
        try {
            out.flush();
        } catch (IOException e) {
            throw new IllegalStateException("Cannot flush outputstream", e);
        }

        // Close resource
        try {
            out.close();
        } catch (IOException e) {
            throw new IllegalStateException("Cannot close outpustream", e);
        }

        IDeployable deployable = getDeployable(file.getAbsolutePath());
        if (deployable instanceof EARDeployable) {
            String parentDir = ((JOnASEARServiceMBean) ServiceManager.getInstance().getEarService()).getAppsDirectory();
            File newFile = new File(parentDir, fileName);
            try {
                FileUtils.copyFile(file, newFile);
            } catch (FileUtilsException e) {
                throw new IllegalStateException("not able to copy '" + file + "' into '" + newFile + "'.", e);
            }
            // Update file
            file = newFile;
        }


        if (!(deployable instanceof RARDeployable)) {
            String modifiedArchive = null;
            ClientGenStubWrapper clientWrapper = new ClientGenStubWrapper();
            try {
                modifiedArchive = clientWrapper.callClientGenStubExecute(file.getPath());
            } catch (Exception e) {
                logger.error("Cannot launch ClientGenStub on the file '" + file + "' : '"
                        + e.getCause().getMessage() + "'.", e);
            }

            // Ugly but required
            try {
                Thread.sleep(1000);
            } catch (InterruptedException e) {
                logger.debug("Cannot sleep ", e);
            }

            boolean modified = false;
            try {
                modified = clientWrapper.callClientGenStubIsInputModifed();
            } catch (Exception e) {
                logger.error("Cannot launch ClientGenStub.isInputModified '" + file + "' : '"
                        + e.getCause().getMessage() + "'.", e);
                // assume file has been modified
                modified = true;
            }

            if (modified && modifiedArchive != null) {
                // assume also that we can call WsGen after
                File newFile = new File(modifiedArchive);
                try {
                    FileUtils.copyFile(newFile, file);
                } catch (Exception e) {
                    logger.error("Error when copying file '" + newFile + "'.", e);
                }

                WsGenWrapper ww = new WsGenWrapper();
                try {
                    modifiedArchive = ww.callWsGenExecute(file.getPath());
                } catch (Exception e) {
                    logger.error("Cannot launch WsGen on the file '" + file + "' : '" + e.getCause().getMessage() + "'.", e);
                }

                try {
                    modified = ww.callWsGenIsInputModifed();
                } catch (Exception e) {
                    logger.error("Cannot launch WsGen on the file '" + file + "' : '"
                            + e.getCause().getMessage() + "'.", e);
                }

                if (modified) {
                    // assume also that we can call WsGen after
                    File newFile2 = new File(modifiedArchive);
                    // Check for modifiedArchive suffix
                    if (modifiedArchive.endsWith(".ear")) {
                        // check if input file is a jar
                        if (deployable instanceof EJBDeployable) {
                            // this case is special: we got a JAR in input and generate a EAR in output
                            // the output goes to apps directory
                            String modifiedDirectory = ((JOnASEARServiceMBean) ServiceManager.getInstance().getEarService()).getAppsDirectory();
                            file = new File(modifiedDirectory + newFile2.getName());
                        }
                    }
                    try {
                        FileUtils.copyFile(newFile2, file);
                    } catch (Exception e) {
                        logger.error("Error when copying file '" + newFile2 + "'. : " + e.getMessage());
                    }
                    return file.getPath();
                }
                return modifiedArchive;
            } else {
                // don't call wsgen as it will fail with DTD too
                return file.getPath();
            }
        } else {
            return file.getPath();
        }

    }

    /**
     * TODO: Remove this method when the other method with new ishmael will work.
     * Deploy file (GenIC), needed for Ishmael to work. The file is in the
     * directory specified by the property "jonas.service.deployment.directory"
     * @param typeparam type of the file (EJB, WAR, EAR, RAR, CAR)
     * @param bfile bytes array of the file
     * @param filename basename of the file to be deployed (if moveIntoDeployableDirectory full path name because local file)
     * @param genicArgs arguments for GenIC
     * @param moveIntoDeployableDirectory true to copy the file into the deployable JOnAS_BASE directory
     * @throws RemoteException
     * @throws EarServiceException
     * @throws ResourceServiceException
     * @throws JWebContainerServiceException
     * @return absolute path of the file
     * @author Dean Jennings. <da_jennings@junta.com.au>
     */
    public String deployFile(Integer typeparam, java.lang.Byte[] bfile, String filename, String[] genicArgs, Boolean moveIntoDeployableDirectory) {
        String directory = "";
        int type = typeparam.intValue();

        ServiceManager serviceManager = null;
        try {
            serviceManager = ServiceManager.getInstance();
        } catch (Exception e) {
            // TODO
            logger.error("Cannot get the instance of the Service Manager : '", e);
        }

        boolean isJar = false;
        if (type == ModuleType.EJB.getValue()) {
                directory = ((JOnASEJBServiceMBean) serviceManager.getEjbService()).getJarsDirectory();
                isJar = true;
        } else if (type == ModuleType.EAR.getValue()) {
                directory = ((JOnASEARServiceMBean) serviceManager.getEarService()).getAppsDirectory();
        } else if (type == ModuleType.WAR.getValue()) {
                directory = ((BaseWebContainerServiceMBean) serviceManager.getWebContainerService()).getWarsDirectory();
        } else if (type == ModuleType.RAR.getValue()) {
                directory = ((JOnASResourceServiceMBean) serviceManager.getResourceService()).getRarsDirectory();
        } else if (type == ModuleType.CAR.getValue()) {
            // do nothing
        }
        File file;
        if (moveIntoDeployableDirectory.booleanValue()) {
           file = new File(directory + filename);
        } else {
           file = new File(filename);
        }
        try {
            FileOutputStream out = new FileOutputStream(file);
            byte[] bfileused = new byte[bfile.length];
            for (int i = 0; i < bfile.length; i++) {
                bfileused[i] = bfile[i].byteValue();
            }
            out.write(bfileused);
            out.close();
            if (type == ModuleType.WAR.getValue()) {
                 checkWebAppDeploymentDesc(file);
            } else {
                if (logger.isDebugEnabled()) {
                    logger.debug("Deployment not yet implemented for type '" + type + ".");
                }
            }
        } catch (IOException ioe) {
            logger.error("Cannot dump to outputstream the file '" + file + "' : '"
                    + ioe.getMessage() + "'.", ioe);
            // TODO
        }

        // Ugly but don't work without it ...
        try {
            Thread.sleep(1000);
        } catch (InterruptedException e1) {
            e1.printStackTrace();
        }

        if (!filename.endsWith(".rar")) {
            String modifiedArchive = null;
            ClientGenStubWrapper clientWrapper = new ClientGenStubWrapper();
            try {
                modifiedArchive = clientWrapper.callClientGenStubExecute(file.getPath());
            } catch (Exception e) {
                logger.error("Cannot launch ClientGenStub on the file '" + file + "' : '"
                        + e.getCause().getMessage() + "'.", e);
            }

            boolean modified = false;
            try {
                modified = clientWrapper.callClientGenStubIsInputModifed();
            } catch (Exception e) {
                logger.error("Cannot launch ClientGenStub.isInputModified '" + file + "' : '"
                        + e.getCause().getMessage() + "'.", e);
                // assume file has been modified
                modified = true;
            }

            if (modified && modifiedArchive != null) {
                // assume also that we can call WsGen after
                File newFile = new File(modifiedArchive);
                try {
                    FileUtils.copyFile(newFile, file);
                } catch (Exception e) {
                    logger.error("Error when copying file '" + newFile + "'.", e);
                }

                // Ugly but don't work without it ...
                try {
                    Thread.sleep(1000);
                } catch (InterruptedException e1) {
                    e1.printStackTrace();
                }

                WsGenWrapper ww = new WsGenWrapper();
                try {
                    modifiedArchive = ww.callWsGenExecute(file.getPath());
                } catch (Exception e) {
                    logger.error("Cannot launch WsGen on the file '" + file + "' : '" + e.getCause().getMessage() + "'.", e);
                }

                try {
                    modified = ww.callWsGenIsInputModifed();
                } catch (Exception e) {
                    logger.error("Cannot launch WsGen on the file '" + file + "' : '"
                            + e.getCause().getMessage() + "'.", e);
                }

                if (modified) {
                    // assume also that we can call WsGen after
                    File newFile2 = new File(modifiedArchive);
                    // Check for modifiedArchive suffix
                    if (modifiedArchive.endsWith(".ear")) {
                        // check if input file is a jar
                        if (isJar) {
                            // this case is special: we got a JAR in input and generate a EAR in output
                            // the output goes to apps directory
                            String modifiedDirectory = ((JOnASEARServiceMBean) serviceManager.getEarService()).getAppsDirectory();
                            file = new File(modifiedDirectory + newFile2.getName());
                        }
                    }
                    try {
                        FileUtils.copyFile(newFile2, file);
                    } catch (Exception e) {
                        logger.error("Error when copying file '" + newFile2 + "'. : " + e.getMessage());
                    }
                    return file.getPath();
                }

                return modifiedArchive;
            } else {
                // don't call wsgen as it will fail with DTD too
                return file.getPath();
            }
        } else {
            return file.getPath();
        }
    }


    /**
     * Validate deployment descriptor of the War file
     * @param file given file
     */
    private void checkWebAppDeploymentDesc(final File file) {
        URL warUrl = null;

        try {
            warUrl = file.toURL();
        } catch (MalformedURLException mue) {
            logger.error("The url for the file '" + file + "' cannot be built : '"
                    + mue.getMessage() + "'.");
            throw new RuntimeException(mue);
        }

        URLClassLoader warCl = new URLClassLoader(new URL[] {warUrl});

        try {
            WebManagerWrapper.getDeploymentDesc(warUrl, warCl, null);
        } catch (WebContainerDeploymentDescException e) {
            String err = "Cannot read the deployment descriptors '" + warUrl.getFile() + "'";
            logger.error(err + ": " + e);
            e.printStackTrace(System.err);
            throw new RuntimeException(err, e);
        }

    }

    /**
     * Send a file to JOnAS Base with the given input stream and write it
     * in the corresponding directory, ie JONAS_BASE/apps, JONAS_BASE/ejbjars, etc.
     * It is based on the filename extension
     * @param fileContent the content (bytes) of the file
     * @param fileName name of the file
     * @param replaceExisting replace existing file if any
     * @return the path of the file
     * @throws Exception if file is already present and if the user don't want to replace existing file.
     */
    public String sendFile(byte[] fileContent, String fileName, boolean replaceExisting) throws Exception {

        File directoryUploadedFile = null;
        FileOutputStream fos = null;
        try {
            // Get directory where to upload file, based on extension
            String dir = getFolderDir(fileName);

            // set the dest file
            directoryUploadedFile = new File(dir, fileName);
            // check, by default we can't overwrite an existing file.
            if (directoryUploadedFile.exists() && !replaceExisting) {
                throw new Exception("File '" + directoryUploadedFile + "' already exists on the server.");
            }

            // write the bytes to the given file
            fos = new FileOutputStream(directoryUploadedFile);
            fos.write(fileContent);
        } finally {
            if (fos != null) {
                try {
                    // close the output stream
                    fos.close();
                } catch (IOException ioe) {
                    logger.debug( "Cannot close the output stream", ioe);
                }
            }

        }
        if (directoryUploadedFile != null) {
            return directoryUploadedFile.getPath();
        } else {
            return "error, no uploaded file";
        }
    }

    /**
     * Remove a specified J2EE module
     * @param fileName Name of file to remove
     * @return true if file has been removed
     * @throws Exception if remove fails
     */
    public boolean removeModuleFile(String fileName) throws Exception {

        // File could exists (absolute file)
        File searchedFile = new File(fileName);
        if (!searchedFile.exists()) {
            // Get directory where to search given file
            String dir = getFolderDir(fileName);
            searchedFile = new File(dir, fileName);
        }
        if (searchedFile.exists()) {
            return searchedFile.delete();
        } else {
            throw new Exception("File '" + searchedFile + "' was not found on the JOnAS server. Cannot remove it");
        }

    }


    /**
     * Get directory based on a filename
     * @param fileName name of the file
     * @return folder of type JONAS_BASE/XXXX/
     * @throws Exception if file is not a J2EE archive
     */
    private String getFolderDir(String fileName) throws Exception {
        String jBase = JProp.getJonasBase();
        // based on extension
        String dir = null;
        // EJB
        if (fileName.toLowerCase().endsWith(".jar")) {
            dir = jBase + File.separator + "ejbjars";
        } else if (fileName.toLowerCase().endsWith(".war")) {
            // War file
            dir = jBase + File.separator + "webapps";
        } else if (fileName.toLowerCase().endsWith(".ear")) {
            // ear file
            dir = jBase + File.separator + "apps";
        } else if (fileName.toLowerCase().endsWith(".rar")) {
            // rar file
            dir = jBase + File.separator + "rars";
        } else {
            // invalid type
            throw new Exception("Invalid extension for the file '" + fileName
                    + "'. Valid are .jar, .war, .ear, .rar");
        }
        return dir;
    }

}
