/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999-2007 Bull S.A.S.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id:RegistryServiceImpl.java 10372 2007-05-14 13:58:42Z sauthieg $
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.registry.carol;

import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import javax.naming.NamingException;

import org.objectweb.carol.jndi.ns.NameServiceManager;
import org.objectweb.carol.util.configuration.CarolDefaultValues;
import org.objectweb.carol.util.configuration.ConfigurationException;
import org.objectweb.carol.util.configuration.ConfigurationRepository;
import org.objectweb.carol.util.configuration.ProtocolConfiguration;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.jonas.lib.execution.ExecutionResult;
import org.ow2.jonas.lib.execution.IExecution;
import org.ow2.jonas.lib.execution.RunnableHelper;
import org.ow2.jonas.lib.service.AbsServiceImpl;
import org.ow2.jonas.lib.util.Log;
import org.ow2.jonas.naming.JComponentContextFactory;
import org.ow2.jonas.naming.JComponentContextFactoryDelegate;
import org.ow2.jonas.registry.RegistryService;
import org.ow2.jonas.registry.carol.delegate.ORBCCFDelegate;
import org.ow2.jonas.service.ServiceException;

/**
 * Implementation of the {@link RegistryService} interface.
 * @author Helene Joanin Contributor(s): Julien Lehembre (Libelis)
 */
public class CarolRegistryService extends AbsServiceImpl implements RegistryService {

    /**
     * Management loggers.
     */
    private static Logger logger = Log.getLogger(Log.JONAS_REGISTRY_PREFIX);

    /**
     * Automatic mode.
     */
    public static final String DEFAULT_MODE = "automatic";

    /**
     * Collocated mode.
     */
    public static final String COLLOCATED = "collocated";

    /**
     * Remote mode.
     */
    public static final String REMOTE = "remote";

    /**
     * Should the local registry be started ?
     */
    private boolean bStartRegistry = true;

    /**
     * Should this service ignore registry startup errors ?
     */
    private boolean bIgnoreError = true;

    /**
     * The ORB Delegate.
     */
    private JComponentContextFactoryDelegate delegate = null;

    /**
     * {@link ComponentContextFactory} factory.
     */
    private JComponentContextFactory factory = null;

    /**
     * server name
     */
    private URL configurationURL;

    /**
     * Assigned to true if we are in OSGi mode
     */
    // TODO Remove it in full OSGi mode
    private boolean isOSGi;

    /**
     * @param mode Registry mode (default, collocated or remote).
     */
    public void setMode(final String mode) {
        if (DEFAULT_MODE.equalsIgnoreCase(mode)) {
            setStartRegistry(true);
            setIgnoreError(true);
        } else if (COLLOCATED.equalsIgnoreCase(mode)) {
            setStartRegistry(true);
            setIgnoreError(false);
        } else if (REMOTE.equalsIgnoreCase(mode)) {
            setStartRegistry(false);
            setIgnoreError(false);
        }
    }

    /**
     * {@inheritDoc}
     */
    public void doStart() throws ServiceException {
        logger.log(BasicLevel.DEBUG, "Starting Carol Registry Service");

        // Configure Carol in an execution block
        IExecution<Void> carolStarter = new IExecution<Void>() {
            public Void execute() throws Exception {
                // Configure Carol
                if (configurationURL != null) {
                    try {
                        ConfigurationRepository.init(configurationURL, getDomainName(), getJonasServerName());
                    } catch (ConfigurationException e) {
                        throw new ServiceException("Cannot init Carol", e);
                    }
                } else {
                    ConfigurationRepository.init(getDomainName(), getJonasServerName());
                }
                return null;
            }
        };

        // Execute
        ExecutionResult<Void> result = RunnableHelper.execute(getClass().getClassLoader(), carolStarter);

        // Throw an ServiceException if needed
        if (result.hasException()) {
            throw new ServiceException(result.getException().getMessage(), result.getException());
        }

        if (bStartRegistry) {
            try {
                if (bIgnoreError) {
                    // start the carol name service manager
                    NameServiceManager.startNonStartedNS();
                } else {
                    NameServiceManager.startNS();
                }
            } catch (Exception e) {
                throw new ServiceException("JOnAS: Cannot start the registry", e);
            }
        }

        // Add delegate
        // TODO Remove this in full OSGi context
        if (!isOSGi) {
            try {
                // Should be done by the container
                setDelegate(new ORBCCFDelegate());
                factory.addDelegate(delegate);
            } catch (NamingException e) {
                throw new ServiceException("Cannot add the delegate", e);
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public void doStop() throws ServiceException {
        logger.log(BasicLevel.DEBUG, "Stopping Carol Registry Service");

        try {
            // Remove the delegate
            // TODO Remove this in full OSGi context
            if (!isOSGi) {
                try {
                    factory.removeDelegate(delegate);
                } catch (NamingException e) {
                    throw new ServiceException("Cannot remove the delegate", e);
                }
            }

            // Stop the carol name service manager
            NameServiceManager.stopNS();
        } catch (Exception e) {
            e.printStackTrace();
            throw new ServiceException("Problem when stopping registry service", e);
        }
    }

    /**
     * {@inheritDoc}
     * @see org.ow2.jonas.lib.service.AbsServiceImpl#checkRequirements()
     */
    @Override
    public void checkRequirements() throws ServiceException {
        // Remove the delegate
        // TODO Remove it in full OSGi mode
        if (!isOSGi) {
            if (factory == null) {
                throwRequirementException("Missing reference on " + JComponentContextFactory.class);
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public List<String> getActiveProtocolNames() {
        List<String> protocols = new ArrayList<String>();

        ProtocolConfiguration[] configs = ConfigurationRepository.getConfigurations();
        for (int i = 0; i < configs.length; i++) {
            protocols.add(configs[i].getName());
        }
        return protocols;
    }

    /**
     * {@inheritDoc}
     */
    public String getDefaultProtocolName() {
        return ConfigurationRepository.getDefaultConfiguration().getName();
    }

    /**
     * {@inheritDoc}
     */
    public int getExportedObjectPort(final String protocolName) {
        ProtocolConfiguration config = getProtocolConfiguration(protocolName);
        String key = CarolDefaultValues.CAROL_PREFIX + protocolName + ".server.port";
        String port = config.getProperties().getProperty(key, "0");
        return Integer.valueOf(port);
    }

    /**
     * {@inheritDoc}
     */
    public String getInitialContextFactoryName(final String protocolName) {
        ProtocolConfiguration config = getProtocolConfiguration(protocolName);
        return config.getProtocol().getInitialContextFactoryClassName();
    }

    /**
     * {@inheritDoc}
     */
    public URI getProviderURL(final String protocolName) {
        ProtocolConfiguration config = getProtocolConfiguration(protocolName);
        return URI.create(config.getProviderURL());
    }

    /**
     * {@inheritDoc}
     */
    public void setDefaultProtocol(final String protocolName) {
        ProtocolConfiguration config = getProtocolConfiguration(protocolName);
        ConfigurationRepository.setCurrentConfiguration(config);
    }

    /**
     * @param protocolName protocol name.
     * @return Returns the {@link Protocol} with the given name.
     */
    private ProtocolConfiguration getProtocolConfiguration(final String protocolName) {
        return ConfigurationRepository.getConfiguration(protocolName);
    }

    /**
     * Bind the {@link JComponentContextFactory} to use.
     * @param factory selected factory
     */
    public void setComponentContextFactory(final JComponentContextFactory factory) {
        this.factory = factory;
    }

    /**
     * Bind the {@link JComponentContextFactoryDelegate} to use.
     * @param delegate selected delegate
     */
    // TODO Remove it in full OSGi mode
    public void setDelegate(final JComponentContextFactoryDelegate delegate) {
        this.delegate = delegate;
    }

    /**
     * @param startRegistry the bStartRegistry to set
     */
    public void setStartRegistry(final boolean startRegistry) {
        bStartRegistry = startRegistry;
    }

    /**
     * @param ignoreError the bIgnoreError to set
     */
    public void setIgnoreError(final boolean ignoreError) {
        bIgnoreError = ignoreError;
    }

    /**
     * @param configurationURL the configurationURL to set
     */
    public void setConfigurationURL(final URL configurationURL) {
        this.configurationURL = configurationURL;
    }

    /**
     * @param configurationURL the configurationURL to set
     * @throws MalformedURLException throws if the URL is not valid
     */
    public void setConfiguration(final String configurationURL) throws MalformedURLException {
        setConfigurationURL(new URL(configurationURL));
    }
}
