/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999-2007 Bull S.A.S.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or 1any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.security.realm.factory;

import java.io.Serializable;
import java.util.Hashtable;

import javax.management.MalformedObjectNameException;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ObjectName;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.naming.Referenceable;

import org.ow2.jonas.jmx.JmxService;
import org.ow2.jonas.lib.reconfig.ReconfigEmitter;
import org.ow2.jonas.lib.reconfig.XMLConfigurationData;
import org.ow2.jonas.lib.util.JonasObjectName;
import org.ow2.jonas.lib.util.Log;
import org.ow2.jonas.security.SecurityService;
import org.ow2.jonas.security.realm.factory.JResource;
import org.ow2.jonas.security.realm.principals.Group;
import org.ow2.jonas.security.realm.principals.Role;
import org.ow2.jonas.security.realm.principals.User;
import org.ow2.jonas.service.manager.ServiceManager;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * This class implements JOnAS realm factory objects. It is build from an xml
 * file by the security service of JOnAS.
 * @author Florent Benoit
 */
public abstract class AbstractJResource extends ReconfigEmitter implements Serializable, Referenceable, JResourceMBean, JResource {

    /**
     * Name of this resource (realm name).
     */
    private String name = null;

    /**
     * The logger used in JOnAS.
     */
    private static Logger logger = Log.getLogger(Log.JONAS_SECURITY_PREFIX);

    /**
     * Hashtable used to cache the users (speed up search).
     */
    private Hashtable users = null;

    /**
     * Value used as sequence number by reconfiguration notifications.
     */
    private long sequenceNumber;

    /**
     * Reference to the security service.
     */
    private SecurityService securityService = null;

    /**
     * Reference to the jmx service.
     */
    private JmxService jmxService = null;

    /**
     * name of the management domain.
     */
    private String domainName = null;

    /**
     * Constructor.
     * @throws Exception if a service can't be retrieved
     */
    public AbstractJResource() throws Exception {
        users = new Hashtable();
        sequenceNumber = 0;
        try {
            ServiceManager sm = ServiceManager.getInstance();
            securityService = (SecurityService) sm.getSecurityService();
            jmxService = (JmxService) sm.getJmxService();
            domainName = jmxService.getDomainName();
        } catch (Exception e) {
            // no save available
            // no MBeans
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "Cannot get MBean server, client mode ? no jmx registration");
            }
        } catch (Error err) {
            // ServiceManager with JProp class not found
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "No services can be retrieved, client mode ?");
            }
        }

    }

    /**
     * @param name the resource (realm) name
     */
    public void setName(String name) {
        this.name = name;
    }

    /**
     * @return the resource (realm) name
     */
    public String getName() {
        return name;
    }

    /**
     * @return the users table
     */
    public Hashtable getUsers() {
        return users;
    }

    /**
     * Return a sequence number and increase this number.
     * @return a sequence number
     */
    protected long getSequenceNumber() {
        return ++sequenceNumber;
    }

    /**
     * Save the resource configuration.
     */
    public void saveConfig() {

        if (securityService != null) {
            // Retrieve the xml value by the Security service
            String xml = securityService.toXML();
            XMLConfigurationData data = new XMLConfigurationData(xml);

            // Send a reconfiguration notification to the listener MBean
            sendReconfigNotification(getSequenceNumber(), name, data);
            sendSaveNotification(getSequenceNumber(), name);

            // rebind this resource into the jndi
            try {
                Context ictx = new InitialContext();
                ictx.rebind(getName(), this);
                if (logger.isLoggable(BasicLevel.DEBUG)) {
                    logger.log(BasicLevel.DEBUG, "jResource " + getName() + " is rebind into the registry.");
                }
            } catch (NamingException e) {
                logger.log(BasicLevel.ERROR, "Cannot rebind factory to the jndi : '" + e.getMessage() + "'");
            }

        }
    }

    /**
     * @return Returns the logger.
     */
    public static Logger getLogger() {
        return logger;
    }

    /**
     * @param logger The logger to set.
     */
    public static void setLogger(final Logger logger) {
        AbstractJResource.logger = logger;
    }

    /**
     * @param users the users table
     */
    public void setUsers(final Hashtable users) {
        this.users = users;
    }

    /**
     *
     */
    public void clearCache() {
        setUsers(new Hashtable());
    }

    /**
     * Allows for listener objects to receive notifications emitted by
     * this resource.
     * @param arg0 listener object
     * @param arg1 notification filter
     * @param arg2 handback object
     */
    public void addNotificationListener(NotificationListener arg0
            , NotificationFilter arg1, Object arg2) {
        super.addNotificationListener(arg0, arg1, arg2);
    }

    /**
     * Create and register an MBean for the user.
     * @param user the user to be managed
     * @throws MalformedObjectNameException if we could not construct ObjectName for the user
     */
    protected void registerUserMBean(User user) throws MalformedObjectNameException {
        String userName = user.getName();
        ObjectName on = JonasObjectName.user(domainName, getName(), userName);
        registerMBean(user, on);
    }

    /**
     * Unregister MBean associated to a user.
     * @param resourceName
     * @param userName
     * @throws MalformedObjectNameException if MBean is not unregistered
     */
    protected void unregisterUserMBean(String resourceName, String userName) throws MalformedObjectNameException {
        unregisterMBean(JonasObjectName.user(domainName, resourceName, userName));
    }

    /**
     * Create and register an MBean for the group
     * @param group the group to be managed
     * @throws MalformedObjectNameException if MBean is not created
     */
    protected void registerGroupMBean(Group group) throws MalformedObjectNameException {
        String groupName = group.getName();
        ObjectName on = JonasObjectName.group(domainName, getName(), groupName);
        registerMBean(group, on);
    }

    /**
     * Unregister MBean associated to a group
     * @param resourceName
     * @param groupName
     * @throws MalformedObjectNameException could not construct ObjectName for the user
     */
    protected void unregisterGroupMBean(String resourceName, String groupName) throws MalformedObjectNameException {
        unregisterMBean(JonasObjectName.group(domainName, resourceName, groupName));
    }

    /**
     * Create and register an MBean for this role
     * @param role the role to be managed
     * @throws MalformedObjectNameException not construct ObjectName for this role
     */
    protected void registerRoleMBean(Role role) throws MalformedObjectNameException {
        String roleName = role.getName();
        ObjectName on = JonasObjectName.role(domainName, getName(), roleName);
        registerMBean(role, on);
    }

    /**
     *
     * @param resourceName
     * @param roleName
     * @throws MalformedObjectNameException could not construct ObjectName for this role
     */
    protected void unregisterRoleMBean(String resourceName, String roleName) throws MalformedObjectNameException {
        unregisterMBean(JonasObjectName.role(domainName, resourceName, roleName));
    }

    /**
     * Allows the resource to register associated MBeans.
     * @param mbean MBean to register
     * @param on MBean's ObjectName
     */
    protected void registerMBean(Object mbean, ObjectName on) {
        jmxService.registerMBean(mbean, on);
    }

    /**
     * Allows the resource to unregister associated MBeans.
     * @param on MBean's ObjectName
     */
    protected void unregisterMBean(ObjectName on) {
        jmxService.unregisterMBean(on);
    }

}
