/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.service.manager.injector;

import java.io.File;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;

import javax.naming.Context;
import javax.naming.NameClassPair;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;

import org.ow2.jonas.service.ServiceException;
import org.ow2.jonas.service.manager.IConfigurationInjector;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * Inject configuration extracted from <code>jonas.properties</code>.
 * @author Guillaume Sauthier
 */
public class ContextParameterInjector implements IConfigurationInjector {

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(ContextParameterInjector.class);

    /**
     * Configuration container.
     */
    private Context container;

    /**
     * Construct a new Injector.
     * @param configurationContainer configuration context
     */
    public ContextParameterInjector(final Context configurationContainer) {
        this.container = configurationContainer;
    }

    /**
     * {@inheritDoc}
     * @see org.ow2.jonas.service.manager.IConfigurationInjector#inject(java.lang.Object)
     */
    public void inject(final Object instance) throws ServiceException {

        logger.debug("Injecting {0}", instance);

        Map<String, String> options = getConfigurationOptions();

        for (String option : options.keySet()) {

            // Exclude .class attribute
            if (!"class".equals(option)) {
                Method m = findSetter(instance.getClass(), option);
                if (m != null) {
                    injectConfiguredValue(instance, m, options.get(option));
                } else {
                    logger.info("Option ''{0}'' not injected", option);
                }
            }
        }

    }

    /**
     * Inject the String value in the instance through a given method.
     * Convert the value on the fly into the expected type.
     * @param instance Object to be configured
     * @param m setter method
     * @param value value as a String
     */
    private void injectConfiguredValue(final Object instance, final Method m, final String value) {
        logger.debug("Inject value: {0} with {1}", m.getName(), value);

        // determine parameter type
        Class<?> type = m.getParameterTypes()[0];
        Object convertedValue = null;

        // convert String to expected type
        // Expect a valueOf(String) static operation
        if (String.class.equals(type)) {
            // directly use the String
            convertedValue = value;
        } else {

            // Another type
            if (type.isPrimitive()) {
                if (Boolean.TYPE.equals(type)) {
                    convertedValue = Boolean.valueOf(value);
                } else if (Byte.TYPE.equals(type)) {
                    convertedValue = Byte.valueOf(value);
                } else if (Integer.TYPE.equals(type)) {
                    convertedValue = Integer.valueOf(value);
                } else if (Double.TYPE.equals(type)) {
                    convertedValue = Double.valueOf(value);
                } else if (Float.TYPE.equals(type)) {
                    convertedValue = Float.valueOf(value);
                } else if (Long.TYPE.equals(type)) {
                    convertedValue = Long.valueOf(value);
                } else if (Character.TYPE.equals(type)) {
                    // If this is a char, only use the 1st char of the String
                    convertedValue = Character.valueOf(value.charAt(0));
                } else if (Short.TYPE.equals(type)) {
                    convertedValue = Short.valueOf(value);
                } else {
                    logger.error("Unhandled primitive type: " + type);
                }
            } else if (List.class.isAssignableFrom(type)) {
                // only handle String list
                // separator is the ','
                String[] values = value.split(",");
                List<String> injection = new ArrayList<String>();

                // We should have at least 1 real value
                if (!((values.length == 1) && ("".equals(values[0])))) {
                    for (int i = 0; i < values.length; i++) {
                        String part = values[i];
                        injection.add(part.trim());
                    }
                }
                convertedValue = injection;

            } else if(File.class.isAssignableFrom(type)) {
                convertedValue = new File(value);
            } else {
                // Try to convert non primitive type
                try {
                    Method convert = type.getMethod("valueOf", new Class[] {String.class});
                    convertedValue = convert.invoke(null, new Object[] {value});
                } catch (Exception e) {
                    throw new ServiceException("Cannot convert argument '" + value + "' to " + type, e);
                }
            }
        }

        try {
            m.invoke(instance, new Object[] {convertedValue});
        } catch (Exception e) {
            throw new ServiceException("Cannot invoke method '" + m + "'", e);
        }
    }

    /**
     * Find an appropriate setter method given an option name.
     * @param clazz Class to be searched for a setter
     * @param option option name (like <code>multicast.port</code>)
     * @return the setter method or null if none found
     */
    private Method findSetter(final Class<?> clazz, final String option) {

        String setterName = convertOptionName(option);

        int index = 0;
        Method[] methods = clazz.getMethods();
        while(index < methods.length) {
            Method m = methods[index];
            if ((setterName.equals(m.getName())) && (m.getParameterTypes().length == 1)) {
                // Found a matching method name
                // OK, a setter has only 1 param.

                return m;
            }
            index++;
        }

        return null;
    }

    /**
     * Convert the option name to a regular setter name.
     * Example: <code>multicast.port</code> becomes <code>setMulticastPort</code>
     * @param option name
     * @return setter method name
     */
    private String convertOptionName(final String option) {
        // capitalize first letter
        String name = capitalize(option);

        // prefix with "set"
        name = "set" + name;

        // Capitalize "x-y" -> "xY" (same for x.y)
        name = capitalizeChars('.', name);
        name = capitalizeChars('-', name);

        logger.debug("{0} -> {1}", option, name);

        return name;
    }

    /**
     * Process String to transform <code>x.Y</code> into <code>xY</code>.
     * @param ch separator character (usually '.')
     * @param name option name to be transformed
     * @return converted option name
     */
    private String capitalizeChars(final char ch, String name) {

        int index;
        while ((index = name.indexOf(ch)) != -1) {
            // prefix
            String pre = name.substring(0, index);
            String sub = name.substring(index + 1);

            name = pre + capitalize(sub);
        }


        return name;
    }

    /**
     * Capitalize the first char of the given String.
     * @param source String to be capitalized
     * @return capitalized String
     */
    private static String capitalize(final String source) {
        return source.substring(0, 1).toUpperCase() + source.substring(1);
    }

    /**
     * Strip the context properties (remove <code>jonas.service.XXX</code>).
     * @return a Map with stripped property names and their String values.
     */
    private Map<String, String> getConfigurationOptions() {

        Map<String, String> options = new Hashtable<String, String>();

        try {
            // Process all the configuration properties
            NamingEnumeration<NameClassPair> properties = container.list("");
            for (; properties.hasMore();) {

                NameClassPair pair = properties.next();
                String property = pair.getName();

                // Only select the one with ...
                if (property.startsWith("jonas.service.")) {

                    // Property name format
                    // jonas.service.<name>.<property>
                    String option = property.substring("jonas.service.".length() + 1);
                    int index = option.indexOf('.');
                    option = option.substring(index + 1);

                    // Store
                    options.put(option, (String) container.lookup(property));
                }
            }
        } catch(NamingException ne) {
            throw new ServiceException("Unable to retrieve options from Context values", ne);
        }

        logger.debug("Options: {0}", options);

        return options;
    }

}
