/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.service.manager.injector;

import java.lang.reflect.Method;

import org.ow2.jonas.service.ServiceException;
import org.ow2.jonas.service.manager.IConfigurationInjector;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * Inject generic resources in the service.
 * @param <I> Interface type.
 * @author Guillaume Sauthier
 */
public class GenericTypeInjector<I> implements IConfigurationInjector {

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(GenericTypeInjector.class);

    /**
     * I implementation to be injected.
     */
    private I injected;

    /**
     * Java interface to be searched.
     */
    private Class<I> javaInterface;

    /**
     * Construct a GenericTypeInjector.
     * @param javaInterface searched interface
     * @param injected instance to be injected
     */
    public GenericTypeInjector(final Class<I> javaInterface, final I injected) {
        this.javaInterface = javaInterface;
        this.injected = injected;
    }

    /**
     * {@inheritDoc}
     * @see org.ow2.jonas.service.manager.IConfigurationInjector#inject(java.lang.Object)
     */
    public void inject(final Object instance) {

        Method setter = findSetter(instance.getClass(), javaInterface);
        if (setter != null) {
            injectValue(instance, setter, injected);
        }

    }

    /**
     * Inject the given value in the target instance.
     * @param instance Object to be injected.
     * @param method setter method to use for Injection.
     * @param injected injected value.
     */
    protected void injectValue(final Object instance, final Method method, final I injected) {
        boolean accessible = method.isAccessible();
        if (!accessible) {
            method.setAccessible(true);
        }
        try {
            method.invoke(instance, injected);
            logger.debug("Method: {0}.{1} injected.",
                         method.getDeclaringClass().getSimpleName(),
                         method.getName());
        } catch (Exception e) {
            throw new ServiceException("Cannot inject dependency through method: " + method, e);
        } finally {
            method.setAccessible(accessible);
        }
    }

    /**
     * Explore the given class hierarchy for appropriate setter method.
     * @param clazz Class to explore.
     * @param searched searched type
     * @return a matching setter method or null if not found.
     */
    protected Method findSetter(final Class<? extends Object> clazz, final Class<I> searched) {

        Class<?> target = clazz;
        // Iterates over the class hierarchy
        while (target != null) {

            Method[] owned = target.getDeclaredMethods();
            for (int i = 0; i < owned.length; i++) {
                Method method = owned[i];

                // Need a setter method ...
                if ((method.getName().startsWith("set"))
                    && (method.getParameterTypes().length == 1)) {

                    Class<?> type = method.getParameterTypes()[0];

                    // with a parameter of searched type
                    if (searched.isAssignableFrom(type)) {
                        // We found a method that is waiting for a typed instance.
                        return method;
                    }
                }
            }

            // Explore parent
            target = target.getSuperclass();
        }

        // not found
        return null;
    }

}
