/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.service.manager.injector;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;

import org.ow2.jonas.service.Service;
import org.ow2.jonas.service.ServiceException;
import org.ow2.jonas.service.manager.IConfigurationInjector;
import org.ow2.jonas.service.manager.ServiceManager;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * Inject services dependencies in the client Services instances.
 * @author Guillaume Sauthier
 */
public class ServiceDependenciesInjector implements IConfigurationInjector {

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(ServiceDependenciesInjector.class);

    /**
     * Manager.
     */
    private ServiceManager manager;

    /**
     * @param manager Service Manager instance.
     */
    public ServiceDependenciesInjector(final ServiceManager manager) {
        this.manager = manager;
    }

    /**
     * {@inheritDoc}
     * @see org.ow2.jonas.service.manager.IConfigurationInjector#inject(java.lang.Object)
     */
    public void inject(final Object instance) {

        logger.debug("Injecting {0}", instance);

        List<Method> dependencies = findServicesDependencies(instance);

        for (Method method : dependencies) {

            // get the dependency
            // Here this is ultra simple, we have only 1 instance of a given service type in JOnAS.
            Class<?> type = method.getParameterTypes()[0];
            Service dependency = manager.getService(type.asSubclass(Service.class));

            if (dependency != null) {
                // assign
                injectValue(instance, method, dependency);
            } else {
                logger.info("Dependency type: ''{0}'' not found.", type);
            }

        }

    }

    /**
     * Inject the given dependency using the given method into the provided instance.
     * @param instance {@link Service} instance
     * @param method Java reflected method
     * @param dependency Other JOnAS {@link Service} to be injected
     */
    private void injectValue(final Object instance, final Method method, final Service dependency) {

        boolean accessible = method.isAccessible();
        if (!accessible) {
            method.setAccessible(true);
        }
        try {
            method.invoke(instance, dependency);
            logger.debug("Method: {0}.{1} injected.",
                         method.getDeclaringClass().getSimpleName(),
                         method.getName());
        } catch (Exception e) {
            throw new ServiceException("Cannot inject dependency through method: " + method, e);
        } finally {
            method.setAccessible(accessible);
        }


    }

    /**
     * Introspect the instance for setter methods expecting a {@link Service} subclass as a parameter.
     * @param instance explored instance
     * @return a {@link List} of matching methods.
     */
    private List<Method> findServicesDependencies(final Object instance) {

        List<Method> methods = new ArrayList<Method>();
        Class<?> clazz = instance.getClass();

        // Explore the class hierarchy
        while (clazz != null) {

            Method[] owned = clazz.getDeclaredMethods();
            for (int i = 0; i < owned.length; i++) {
                Method method = owned[i];

                // Need a setter method ...
                if ((method.getName().startsWith("set")) && (method.getParameterTypes().length == 1)) {
                    Class<?> type = method.getParameterTypes()[0];

                    // with a parameter of type Service
                    if (Service.class.isAssignableFrom(type)) {
                        // We found a method that is waiting for a Service instance.
                        methods.add(method);
                    }
                }
            }

            // Explore parent
            clazz = clazz.getSuperclass();
        }
        return methods;
    }

}
