/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.tm.jotm.osgi;

import org.osgi.framework.BundleActivator;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceReference;
import org.osgi.framework.ServiceRegistration;
import org.ow2.jonas.jmx.JmxService;
import org.ow2.jonas.lib.execution.ExecutionResult;
import org.ow2.jonas.lib.execution.IExecution;
import org.ow2.jonas.lib.execution.RunnableHelper;
import org.ow2.jonas.naming.JComponentContextFactory;
import org.ow2.jonas.naming.JComponentContextFactoryDelegate;
import org.ow2.jonas.naming.JNamingManager;
import org.ow2.jonas.properties.ServerProperties;
import org.ow2.jonas.tm.TransactionService;
import org.ow2.jonas.tm.jotm.JOTMTransactionService;
import org.ow2.jonas.tm.jotm.delegate.UserTransactionCCFDelegate;

/**
 * JOTM Transaction Service Activator.
 * @author Philippe Durieux
 * TODO Needs a better logging solution :)
 */
public class Activator implements BundleActivator {

    private ServiceRegistration srTms;
    private ServiceRegistration srUtx;
    private ServerProperties props;

    /**
     * The service instance.
     */
    private JOTMTransactionService tms;

    /**
     * Start JOTM Transaction Service Bundle.
     * @param bc OSGi {@link BundleContext}
     * @throws Exception when TS startup has failed.
     * @see org.osgi.framework.BundleActivator#start(org.osgi.framework.BundleContext)
     */
    public void start(final BundleContext bc) throws Exception {
        // TODO log this !
        System.out.println("Starting JOTM Transaction Service Bundle...");

        // Get DomainName and ServerName from the Bootstrap service
        // TODO assume service already started.
        ServiceReference sr = bc.getServiceReference(ServerProperties.class.getName());
        props = (ServerProperties) bc.getService(sr);

        // Start the service in an execution block
        IExecution<JOTMTransactionService> tmsStarter = new IExecution<JOTMTransactionService>() {

            public JOTMTransactionService execute() throws Exception {
                // Create the Service
                tms = new JOTMTransactionService();

                // Configuration
                tms.setServerProperties(props);
                boolean jtmremote = props.getValueAsBoolean("jonas.service.jtm.remote", false);
                tms.setRemote(jtmremote);

                // dependencies
                ServiceReference sr = bc.getServiceReference(JmxService.class.getName());
                if (sr != null) {
                    final JmxService jmx = (JmxService) bc.getService(sr);
                    tms.setJmxService(jmx);
                } else {
                    System.out.println("JmxService not running");
                }
                sr = bc.getServiceReference(JNamingManager.class.getName());
                if (sr != null) {
                    final JNamingManager naming = (JNamingManager) bc.getService(sr);
                    tms.setNaming(naming);
                } else {
                    System.out.println("JNamingManager not running");
                }
                sr = bc.getServiceReference(JComponentContextFactory.class.getName());
                if (sr != null) {
                    final JComponentContextFactory ccf = (JComponentContextFactory) bc.getService(sr);
                    tms.setComponentContextFactory(ccf);
                } else {
                    System.out.println("JNamingManager not running");
                }
                tms.checkRequirements();

                // Start the Service
                tms.start();

                return tms;

            }
        };

        // Execute
        ExecutionResult<JOTMTransactionService> result = RunnableHelper.execute(Activator.class.getClassLoader(), tmsStarter);

        // Throw an Exception if needed
        if (result.hasException()) {
            throw result.getException();
        }
        // Store the ref
        tms = result.getResult();

        // Register the Service to OSGi registry
        srTms = bc.registerService(TransactionService.class.getName(), tms, null);

        // register the delegate too
        UserTransactionCCFDelegate delegate = new UserTransactionCCFDelegate();
        srUtx = bc.registerService(JComponentContextFactoryDelegate.class.getName(), delegate, null);
    }

    /**
     * {@inheritDoc}
     * @see org.osgi.framework.BundleActivator#stop(org.osgi.framework.BundleContext)
     */
    public void stop(final BundleContext bc) throws Exception {
        System.out.println("Stopping JOTM Transaction Service Bundle...");

        srUtx.unregister();
        srTms.unregister();
        tms.stop();
    }

}
