/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.eclipse;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Enumeration;

/**
 * Allows to use a preferred classloader (for JONAS_BASE/conf for example)
 * before asking parent classloader.
 * @author Florent BENOIT
 */
public class DelegatingClassLoader extends URLClassLoader {

    /**
     * Build a new classloader with given set of URLs and Parent Class Loader.
     * @param urls given urls
     * @param parentClassLoader the parent class loader
     */
    public DelegatingClassLoader(final URL[] urls, final ClassLoader parentClassLoader) {
        super(urls, parentClassLoader);
    }

    /**
     * Finds the resource with the given name. A resource is some data (images,
     * audio, text, etc) that can be accessed by class code in a way that is
     * independent of the location of the code.
     * @param name The resource name
     * @return A <tt>URL</tt> object for reading the resource, or
     *         <tt>null</tt> if the resource could not be found or the invoker
     *         doesn't have adequate privileges to get the resource.
     */
    public URL getResource(final String name) {
        // Search local first
        URL url;
        url = findResource(name);
        if (url != null) {
            return url;
        }
        return super.getResource(name);
    }

    /**
     * Returns an input stream for reading the specified resource.
     * @param name The resource name
     * @return An input stream for reading the resource, or <tt>null</tt> if
     *         the resource could not be found
     */
    public InputStream getResourceAsStream(final String name) {
        // Search local first
        URL url = getResource(name);
        try {
            if (url != null) {
                return url.openStream();
            }
        } catch (IOException e) {
            return null;
        }
        return null;
    }

    /**
     * Finds all the resources with the given name.
     * @param name The resource name
     * @return An enumeration of {@link java.net.URL <tt>URL</tt>} objects for
     *         the resource. If no resources could be found, the enumeration
     *         will be empty. Resources that the class loader doesn't have
     *         access to will not be in the enumeration.
     * @throws IOException If I/O errors occur
     */
    public Enumeration<URL> getResources(final String name) throws IOException {
        // Search local first
        Enumeration<URL> urls = findResources(name);
        if (urls != null) {
            return urls;
        }
        return super.getResources(name);
    }

    /**
     * Loads the class with the specified <a href="#name">binary name</a>. This
     * method searches for classes in the same manner as the {@link
     * #loadClass(String, boolean)} method. It is invoked by the Java virtual
     * machine to resolve class references. Invoking this method is equivalent
     * to invoking
     * {@link #loadClass(String, boolean) <tt>loadClass(name, false)</tt>}.
     * @param name The <a href="#name">binary name</a> of the class
     * @return The resulting <tt>Class</tt> object
     * @throws ClassNotFoundException If the class was not found
     */
    public Class<?> loadClass(final String name) throws ClassNotFoundException {
        return loadClass(name, false);
    }

    /**
     * Loads the class with the specified <a href="#name">binary name</a>. The
     * default implementation of this method searches for classes in the
     * following order:
     * @param name The <a href="#name">binary name</a> of the class
     * @param resolve If <tt>true</tt> then resolve the class
     * @return The resulting <tt>Class</tt> object
     * @throws ClassNotFoundException If the class could not be found
     */
    protected synchronized Class<?> loadClass(final String name, final boolean resolve) throws ClassNotFoundException {

        // The JDK classes (and Monolog) are loaded with the system classloader
        if (name.startsWith("java.") || name.startsWith("sun.") || name.startsWith("com.sun.") || name.startsWith("javax.")
                || name.startsWith("org.ietf.") || name.startsWith("org.omg.") || name.startsWith("org.w3c.")
                || name.startsWith("org.xml.") || name.startsWith("org.objectweb.util.monolog")) {
                return super.loadClass(name, resolve);
        }

        // First, check if the class has already been loaded
        Class<?> c = findLoadedClass(name);

        // Not loaded, find the class
        if (c == null) {
            c = findClass(name);
        }

        // Needs to resolve the class ?
        if (resolve) {
            resolveClass(c);
        }

        // Found, return it
        if (c != null) {
            return c;
        }

        // Ask parent classloader
        return super.loadClass(name, resolve);
    }

}
