/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.eclipse;

import java.io.File;
import java.lang.reflect.Method;
import java.util.List;
import java.util.Properties;

import org.ow2.fastrmic.ClassInfo;
import org.ow2.fastrmic.RMIC;

/**
 * Allow to start JOnAS within eclipse without creating any assembly.<br>
 * This class is for JOnAS developers.<br>
 * NOTE : The JONAS_BASE/conf folder needs to be added before launching this
 * class.
 * @author Florent BENOIT
 */
public final class JOnASEclipse {

    /**
     * Utility class, no public constructor.
     */
    private JOnASEclipse() {

    }

    /**
     * Server main routine. No args.
     * @param args the arguments of the server
     * @exception Exception can not start server
     */
    public static void main(final String[] args) throws Exception {

        // Set JOnAS Root to the root folder
        System.setProperty("install.root", ".");

        // Set JOnAS BASE
        String jonasBase = System.getProperty("jonas.base");

        if (jonasBase == null) {
            System.err.println("JONAS_BASE is not set");
            System.exit(0);
        }

        File jBase = new File(jonasBase);
        if (!jBase.exists()) {
            System.err.println("JONAS_BASE '" + jBase + "' doesn't exists");
            System.exit(0);
        }

        System.out.println("USING JONAS_BASE = " + jonasBase);


        // Set other properties
        System.setProperty("java.awt.headless", "true");
        System.setProperty("java.security.policy", jonasBase + File.separator + "conf" + File.separator + "java.policy");
        System.setProperty("java.security.auth.login.config", jonasBase + File.separator + "conf" + File.separator
                + "jaas.config");
        System.setProperty("jonas.classpath", "");
        System.setProperty("jonas.default.classloader", "true");
        System.setProperty("org.omg.CORBA.ORBClass", "org.jacorb.orb.ORB");
        System.setProperty("org.omg.CORBA.ORBSingletonClass", "org.jacorb.orb.ORBSingleton");
        System.setProperty("org.omg.PortableInterceptor.ORBInitializerClass.standard_init",
                "org.jacorb.orb.standardInterceptors.IORInterceptorInitialize");
        System.setProperty("javax.rmi.CORBA.PortableRemoteObjectClass", "org.objectweb.carol.rmi.multi.MultiPRODelegate");
        System.setProperty("java.naming.factory.initial", "org.objectweb.carol.jndi.spi.MultiOrbInitialContextFactory");
        System.setProperty("javax.rmi.CORBA.UtilClass", "org.objectweb.carol.util.delegate.UtilDelegateImpl");

        // Endorsed ?
        // System.setProperty("java.endorsed.dirs=$JONAS_ROOT/lib/endorsed");

        // Use our classloader for now
        ClassLoader currentCL = Thread.currentThread().getContextClassLoader();

        // Get instance of loader manager
        Class<?> loaderManagerClass = currentCL.loadClass("org.ow2.jonas.lib.bootstrap.LoaderManager");
        Method lmGetInstanceMeth = loaderManagerClass.getMethod("getInstance");
        Object lm = lmGetInstanceMeth.invoke(null);

        // Get an instance of JProp
        Class<?> jPropClass = currentCL.loadClass("org.ow2.jonas.lib.bootstrap.JProp");
        Method jPgetInstanceMeth = jPropClass.getMethod("getInstance");
        Object jProp = jPgetInstanceMeth.invoke(null);

        // Init loader manager
        Method initMethod = loaderManagerClass.getMethod("init", jPropClass);
        initMethod.invoke(lm, jProp);

        // Load Server class
        Class<?> clazz = currentCL.loadClass("org.ow2.jonas.server.Server");
        Method m = clazz.getMethod("main", String[].class);

        // Generate stubs
        genStub("org.ow2.jonas.security.realm.factory.JResourceRemoteImpl");
        genStub("org.ow2.jonas.server.Adm");

        // Get JProp properties
        Method getEnvMethod = jPropClass.getMethod("getEnv");
        Properties p = (Properties) getEnvMethod.invoke(jProp);

        p.put("jonas.service.ejb.auto-genic", "false");
        p.put("jonas.service.ws.auto-wsgen.engaged", "false");

        // Invoke class
        m.invoke(null, new Object[] {args});

    }

    /**
     * Generates a stub for the given classname.
     * @param className the name of the class
     * @throws Exception if the stub is not generated
     */
    public static void genStub(final String className) throws Exception {
        List<ClassInfo> classInfos = RMIC.generateStubForClass(className, Thread.currentThread().getContextClassLoader());
        for (ClassInfo classInfo : classInfos) {
            defineClass(classInfo, Thread.currentThread().getContextClassLoader());
        }
    }

    /**
     * Defines a class in the current classloader.
     * @param classInfo the data for the class
     * @param classLoader the classloader to use
     * @throws Exception if the class is not defined
     */
    public static void defineClass(final ClassInfo classInfo, final ClassLoader classLoader) throws Exception {

        Method defineClassMethod = ClassLoader.class.getDeclaredMethod("defineClass", new Class[] {String.class, byte[].class,
                int.class, int.class });

        // protected method invocaton
        defineClassMethod.setAccessible(true);
        Object[] args = new Object[] {classInfo.getName(), classInfo.getBytecode(), Integer.valueOf(0),
                new Integer(classInfo.getBytecode().length) };
        try {
            defineClassMethod.invoke(classLoader, args);
        } finally {
            defineClassMethod.setAccessible(false);
        }
    }
}
