/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or 1any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.launcher.felix;

import java.io.File;
import java.io.IOException;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;

import org.apache.felix.framework.cache.BundleCache;
import org.apache.felix.framework.util.FelixConstants;
import org.ow2.jonas.launcher.felix.util.Maven2Utils;

/**
 * Default {@link IConfigurationProvider} for Apache Felix.
 * @author Guillaume Sauthier
 */
public class DefaultConfigurationProvider implements IConfigurationProvider {

    /**
     * @return a Felix default configuration.
     * @throws IOException configuration not found
     */
    @SuppressWarnings("unchecked")
    public Map getConfiguration() throws IOException {

        // Get the default configuration files
        Properties config = loadDefaultConfiguration();

        // Augment the basis configuration
        findAutoDeployedBundles(config);

        // Place the cache directory in ${user.home}/.jonas
        // TODO should be configurable
        String userHome = System.getProperty("user.home");
        File home = new File(userHome, ".jonas");

        // Set the cache directory
        config.put(BundleCache.CACHE_PROFILE_DIR_PROP, home.getAbsolutePath());
        return config;
    }

    /**
     * Update the given config with start level options.
     * @param config configuration to enhance.
     * @throws IOException if default property file is not found.
     */
    private void findAutoDeployedBundles(final Properties config) throws IOException {

        // Get property file content
        Properties levels = getAutoDeployedBundles();

        // Iterates over the properties
        for (Iterator<Object> it = levels.keySet().iterator(); it.hasNext();) {

            // Only process start level properties
            String key = (String) it.next();
            if (key.startsWith("start.level.")) {

                // Get the associated level
                String level = key.substring("start.level".length());
                String value = levels.getProperty(key);

                StringBuilder sb = new StringBuilder();

                // The bundle list separator is ','
                String[] bundles = value.split(",");
                for (int i = 0; i < bundles.length; i++) {
                    String bundle = bundles[i];

                    // Only process non-empty parts
                    if (!"".equals(bundle)) {

                        // The bundle is specified using the following format:
                        // <groupId>:<artifactId>[:<version>]
                        String[] artifact = bundle.split(":");

                        String groupId = artifact[0].trim();
                        String artifactId = artifact[1].trim();

                        // if version is not provided, use JOnAS' one.
                        String version;
                        if (artifact.length > 2) {
                            version = artifact[2].trim();
                        } else {
                            version = getJOnASVersion();
                        }

                        sb.append(Maven2Utils.getBundleMaven2Location(groupId, artifactId, version).concat(" "));
                    }
                }

                // Creates the really property
                config.put(FelixConstants.AUTO_START_PROP.concat(level), sb.toString());
            }
        }
    }

    /**
     * @return the config file for auto deployed bundles.
     * @throws IOException file not found
     */
    private static Properties getAutoDeployedBundles() throws IOException {
        return getPropertiesResource("jonas-autodeploy-bundles.properties");
    }

    /**
     * @return the default set of Felix properties.
     * @throws IOException file not found
     */
    private static Properties loadDefaultConfiguration() throws IOException {
        return getPropertiesResource("default-config.properties");
    }

    /**
     * @param resource properties file name (relative to org/ow2/jonas/launcher/felix)
     * @return the content of the resource as a {@link Properties} instance.
     * @throws IOException if file is not found or if file is not a property file.
     */
    private static Properties getPropertiesResource(final String resource) throws IOException {
        Properties config = new Properties();
        java.io.InputStream is = JOnAS.class.getResourceAsStream(resource);
        config.load(is);
        return config;
    }

    /**
     * @return the JOnAS version from the Implementation-Version Manifest entry.
     */
    private static String getJOnASVersion() {
        return JOnAS.class.getPackage().getImplementationVersion();
    }

}
